# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024, 2025, 2026 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA REST API client."""

import json
import logging
import os
import traceback
from functools import partial
from urllib.parse import urljoin

import requests
from bravado.exception import HTTPError
from reana_client.api.utils import get_content_disposition_filename
from reana_client.config import ERROR_MESSAGES
from reana_client.errors import FileDeletionError, FileUploadError
from reana_client.utils import is_regular_path, is_uuid_v4
from reana_commons.api_client import get_current_api_client
from reana_commons.config import REANA_WORKFLOW_ENGINES
from reana_commons.errors import REANASecretAlreadyExists, REANASecretDoesNotExist
from reana_commons.specification import (
    load_input_parameters,
    load_workflow_spec_from_reana_yaml,
)
from reana_commons.validation.utils import validate_reana_yaml, validate_workflow_name
from werkzeug.local import LocalProxy

current_rs_api_client = LocalProxy(
    partial(get_current_api_client, component="reana-server")
)


def ping(access_token):
    """Check if the REANA server is reachable and the user is correctly authenticated.

    :param access_token: access token of the current user.

    :return: a dictionary with the ``status`` key (``"Connected"`` if the server is reachable, the error message if
             there is a problem), the ``error`` key (``True`` if there is an error, ``False`` otherwise),
             and info about the current user in ``full_name`` and ``email``.
    """
    try:
        response, http_response = current_rs_api_client.api.get_you(
            access_token=access_token
        ).result()
        if http_response.status_code == 200:
            response["status"] = "Connected"
            response["error"] = False
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "REANA server health check failed: "
            "\nStatus: {}\nReason: {}".format(e.response.status_code, e.response.reason)
        )
        if e.response.status_code == 404:
            return {"status": "ERROR: INVALID SERVER", "error": True}
        if e.response.status_code == 403:
            return {"status": "ERROR: INVALID ACCESS TOKEN", "error": True}
        raise Exception(e.response)
    except Exception:
        return {"status": "ERROR: INVALID SERVER", "error": True}


def get_user_quota(access_token):
    """Retrieve user quota usage and limits.

    :param access_token: access token of the current user.

    :return: a dictionary with the information about the usage and limits of the user's quota.
             The keys are ``cpu`` and ``disk``, and refer to the respective usage and limits.
    """
    try:
        response, http_response = current_rs_api_client.api.get_you(
            access_token=access_token
        ).result()
        if http_response.status_code == 200:
            return response["quota"]
        raise Exception(
            "Expected status code 200 but replied with "
            "{status_code}".format(status_code=http_response.status_code)
        )

    except HTTPError as e:
        logging.debug(
            "User quotas could not be retrieved: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def get_workflows(
    access_token,
    type,
    verbose=False,
    page=None,
    size=None,
    status=None,
    search=None,
    include_progress=None,
    include_workspace_size=None,
    workflow=None,
    shared=None,
    shared_by=None,
    shared_with=None,
):
    """List all existing workflows.

    :param access_token: access token of the current user.
    :param type: type of workflow to be listed: ``"interactive"`` if you want to
                 list only the workflows that have an interactive session attached, with the
                 info about the session, or ``"batch"`` (default) otherwise.
    :param verbose: show detailed information about workflows.
    :param page: page number of the paginated list of workflows.
    :param size: number of workflows per page.
    :param status: filter workflows by status.
    :param search: search workflows by name.
    :param include_progress: include progress information in the response.
    :param include_workspace_size: include workspace size information in the response.
    :param workflow: name or id of the workflow.
    :param shared: list all shared (owned and unowned) workflows.
    :param shared_by: list workflows shared by the specified user(s).
    :param shared_with: list workflows shared with the specified user(s).

    :return: a list of dictionaries with the information about the workflows.
             The information includes the workflow ``name``, ``id``, ``status``, ``size``,
             ``user`` (given as the user's ID), and info about the interactive session if
             present.
    """
    try:
        response, http_response = current_rs_api_client.api.get_workflows(
            access_token=access_token,
            verbose=verbose,
            type=type,
            page=page,
            size=size,
            status=status,
            search=search,
            include_progress=include_progress,
            include_workspace_size=include_workspace_size,
            workflow_id_or_name=workflow,
            shared=shared,
            shared_by=shared_by,
            shared_with=shared_with,
        ).result()
        if http_response.status_code == 200:
            return response.get("items")
        raise Exception(
            "Expected status code 200 but replied with "
            "{status_code}".format(status_code=http_response.status_code)
        )

    except HTTPError as e:
        logging.debug(
            "The list of workflows could not be retrieved: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def get_workflow_status(workflow, access_token):
    """Get status of previously created workflow.

    :param workflow: name or id of the workflow.
    :param access_token: access token of the current user.

    :return: a dictionary with the information about the workflow status.
             The dictionary has the following keys: ``id``, ``logs``, ``name``,
             ``progress``, ``status``, ``user``.
    """
    try:
        response, http_response = current_rs_api_client.api.get_workflow_status(
            workflow_id_or_name=workflow, access_token=access_token
        ).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Analysis status could not be retrieved: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def create_workflow(reana_specification, name, access_token):
    """Create a workflow.

    :param reana_specification: a dictionary representing the REANA specification of the workflow.
    :param name: name of the workflow.
    :param access_token: access token of the current user.

    :return: if the workflow was created successfully, a dictionary with the information about
             the ``workflow_id`` and ``workflow_name``, along with a ``message`` of success.
    """
    try:
        response, http_response = current_rs_api_client.api.create_workflow(
            reana_specification=json.loads(
                json.dumps(reana_specification, sort_keys=True)
            ),
            workflow_name=name,
            access_token=access_token,
        ).result()
        if http_response.status_code == 201:
            return response
        else:
            raise Exception(
                "Expected status code 201 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Workflow creation failed: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def create_workflow_from_json(
    name,
    access_token,
    workflow_json=None,
    workflow_file=None,
    parameters=None,
    workflow_engine="yadage",
    outputs=None,
    workspace_path=None,
):
    """Create a workflow from JSON specification.

    :param name: name or UUID of the workflow to be started.
    :param access_token: access token of the current user.
    :param workflow_json: workflow specification in JSON format.
    :param workflow_file: workflow specification file path.
                          Ignores ``workflow_json`` if provided.
    :param parameters: workflow input parameters dictionary.
    :param workflow_engine: one of the workflow engines (yadage, serial, cwl)
    :param outputs: dictionary with expected workflow outputs.
    :param workspace_path: path to the workspace where the workflow is located.

    :return: if the workflow was created successfully, a dictionary with the information about
             the ``workflow_id`` and ``workflow_name``, along with a ``message`` of success.

    :Example:

      .. code:: python

        create_workflow_from_json(
            workflow_json=workflow_json,
            name='workflow_name.1',
            access_token='access_token',
            parameters={'files': ['file.txt'],
                'parameters': {'key': 'value'}},
            workflow_engine='serial')
    """
    validate_workflow_name(name)
    if is_uuid_v4(name):
        raise ValueError("Workflow name cannot be a valid UUIDv4")
    if not access_token:
        raise Exception(ERROR_MESSAGES["missing_access_token"])
    if os.environ.get("REANA_SERVER_URL") is None:
        raise Exception("Environment variable REANA_SERVER_URL is not set")
    workflow_engine = workflow_engine.lower()
    if workflow_engine not in REANA_WORKFLOW_ENGINES:
        raise Exception(
            "Workflow engine - {} not found. You must use one of "
            "these engines - {}".format(workflow_engine, REANA_WORKFLOW_ENGINES)
        )
    try:
        reana_yaml = dict(workflow={})
        reana_yaml["workflow"]["type"] = workflow_engine
        if parameters:
            reana_yaml["inputs"] = parameters
        if outputs:
            reana_yaml["outputs"] = outputs
        if workflow_file:
            reana_yaml["workflow"]["file"] = workflow_file
            reana_yaml["workflow"]["specification"] = (
                load_workflow_spec_from_reana_yaml(reana_yaml, workspace_path)
            )
        else:
            reana_yaml["workflow"]["specification"] = workflow_json
        # The function below loads the input parameters into the reana_yaml dictionary
        # taking them from the parameters yaml files (used by CWL and Snakemake workflows),
        # and replacing the `input.parameters.input` field with the actual parameters values.
        # For this reason, we have to load the workflow specification first, as otherwise
        # the specification validation would fail.
        input_params = load_input_parameters(reana_yaml, workspace_path)
        if input_params is not None:
            reana_yaml["inputs"]["parameters"] = input_params
        validate_reana_yaml(reana_yaml)
        response, http_response = current_rs_api_client.api.create_workflow(
            reana_specification=json.loads(json.dumps(reana_yaml, sort_keys=True)),
            workflow_name=name,
            access_token=access_token,
        ).result()
        if http_response.status_code == 201:
            return response
        else:
            raise Exception(
                "Expected status code 201 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Workflow creation failed: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def start_workflow(workflow, access_token, parameters):
    """Start a workflow.

    :param workflow: name or id of previously created workflow.
    :param access_token: access token of the current user.
    :param parameters: dict of workflow parameters to override the original
        ones (after workflow creation).

    :return: if the workflow was started successfully, a dictionary with the information about
             the ``workflow_id``, ``workflow_name``, ``run_number``, ``status``, and ``user``,
             along with a ``message`` of success.
    """
    try:
        response, http_response = current_rs_api_client.api.start_workflow(
            workflow_id_or_name=workflow,
            access_token=access_token,
            parameters=parameters,
        ).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Workflow run could not be started: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def upload_file(workflow, file_, file_name, access_token):
    """Upload file to workflow workspace.

    :param workflow: name or id of the workflow.
    :param file_: content of a file that will be uploaded.
    :param file_name: name of a file that will be uploaded.
    :param access_token: access token of the current user.

    :return: if the file was uploaded successfully, a dictionary
             with a ``message`` of success.
    """
    from reana_client.utils import get_api_url

    try:
        endpoint = current_rs_api_client.api.upload_file.operation.path_name.format(
            workflow_id_or_name=workflow
        )
        http_response = requests.post(
            urljoin(get_api_url(), endpoint),
            data=file_,
            params={"file_name": file_name, "access_token": access_token},
            headers={"Content-Type": "application/octet-stream"},
            verify=False,
        )
        if http_response.ok:
            return http_response.json()
        raise Exception(http_response.json().get("message"))
    except requests.exceptions.ConnectionError:
        logging.debug("File could not be uploaded.", exc_info=True)
        raise Exception("Could not connect to the server {}".format(get_api_url()))
    except requests.exceptions.HTTPError as e:
        logging.debug("The server responded with an HTTP error code.", exc_info=True)
        raise Exception("Unexpected response from the server: \n{}".format(e.response))
    except requests.exceptions.Timeout:
        logging.debug("Timeout while trying to establish connection.", exc_info=True)
        raise Exception("The request to the server has timed out.")
    except requests.exceptions.RequestException:
        logging.debug(
            "Something went wrong while connecting to the server.", exc_info=True
        )
        raise Exception("The request to the server has failed for an unknown reason.")
    except Exception as e:
        raise e


def get_workflow_logs(workflow, access_token, steps=None, page=None, size=None):
    """Get logs from a workflow engine.

    :param workflow: name or id of the workflow.
    :param access_token: access token of the current user.
    :param steps: list of step names to get logs for.
    :param page: page number of returned log list.
    :param size: page size of returned log list.

    :return: a dictionary with a ``logs`` key containing a JSON string that
             contains the requested logs.
    """
    try:
        response, http_response = current_rs_api_client.api.get_workflow_logs(
            workflow_id_or_name=workflow,
            steps=steps,
            access_token=access_token,
            page=page,
            size=size,
        ).result()

        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Workflow logs could not be retrieved: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def download_file(workflow, file_name, access_token):
    """Download the requested file if it exists.

    :param workflow: name or id of the workflow.
    :param file_name: file name or path to the file requested.
    :param access_token: access token of the current user.

    :return: a tuple containing file binary content, filename and whether
        the returned file is a zip archive containing multiple files.
    """
    try:
        from reana_client.utils import get_api_url

        logging.getLogger("urllib3").setLevel(logging.CRITICAL)
        endpoint = current_rs_api_client.api.download_file.operation.path_name.format(
            workflow_id_or_name=workflow, file_name=file_name
        )
        http_response = requests.get(
            urljoin(get_api_url(), endpoint),
            params={"file_name": file_name, "access_token": access_token},
            verify=False,
        )
        if "Content-Disposition" in http_response.headers:
            file_name = get_content_disposition_filename(
                http_response.headers.get("Content-Disposition")
            )

        # A zip archive is downloaded if multiple files are requested
        multiple_files_zipped = (
            http_response.headers.get("Content-Type") == "application/zip"
        )

        if http_response.status_code == 200:
            return http_response.content, file_name, multiple_files_zipped
        else:
            raise Exception(
                "Error {status_code} {reason} {message}".format(
                    status_code=http_response.status_code,
                    reason=http_response.reason,
                    message=http_response.json().get("message"),
                )
            )

    except HTTPError as e:
        logging.debug(
            "Output file could not be downloaded: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def delete_file(workflow, file_name, access_token):
    """Delete the requested file if it exists.

    :param workflow: name or id of the workflow.
    :param file_name: file name or path to the file requested.
    :param access_token: access token of the current user.

    :return: a dictionary with two keys: ``deleted`` and ``failed``.
             Each of this keys contains another dictionary with the
             name of the file as key and info about the size as value.
    """
    try:
        response, http_response = current_rs_api_client.api.delete_file(
            workflow_id_or_name=workflow,
            file_name=file_name,
            access_token=access_token,
        ).result()
        if http_response.status_code == 200 and (
            response["deleted"] or response["failed"]
        ):
            return response
        elif not (response["deleted"] or response["failed"]):
            raise FileDeletionError(
                "{} did not match any existing " "file.".format(file_name)
            )
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "File could not be downloaded: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def list_files(
    workflow, access_token, file_name=None, page=None, size=None, search=None
):
    """Return the list of files for a given workflow workspace.

    :param workflow: name or id of the workflow.
    :param access_token: access token of the current user.
    :param file_name: file name(s) (glob) to list.
    :param page: page number of returned file list.
    :param size: page size of returned file list.
    :param search: filter search results by parameters.
    :returns: a list of dictionaries that have the ``name``, ``size`` and
                ``last-modified`` keys.
    """
    try:
        response, http_response = current_rs_api_client.api.get_files(
            workflow_id_or_name=workflow,
            access_token=access_token,
            file_name=file_name,
            page=page,
            size=size,
            search=search,
        ).result()

        if http_response.status_code == 200:
            return response.get("items")
        raise Exception(
            "Expected status code 200 but replied with "
            "{status_code}".format(status_code=http_response.status_code)
        )

    except HTTPError as e:
        logging.debug(
            "File list could not be retrieved: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def upload_to_server(workflow, paths, access_token):
    """Upload file or directory to REANA server.

    Shared e.g. by `code upload` and `inputs upload`.

    :param workflow: name or id of workflow whose workspace should be
        used to store the files.
    :param paths: absolute filepath(s) of files to be uploaded.
    :param access_token: access token of the current user.

    :return: the list of path of files that were uploaded.
    """
    if not workflow:
        raise ValueError("Workflow name or id must be provided")
    if not paths:
        logging.info(
            "No path(s) to file(s) that should be uploaded to workspace was provided."
        )
        return []

    logging.info('Workflow "{}" selected'.format(workflow))

    # Check if multiple paths were given and iterate over them
    if type(paths) is list or type(paths) is tuple:
        for path in paths:
            upload_to_server(workflow, path, access_token)
    # `paths` points to a single file or directory
    else:
        path = paths
        if ".." in paths.split("/"):
            raise FileUploadError('Path cannot contain ".."')

        if not is_regular_path(path):
            logging.info(f"Ignoring symlink {path}")
            return []

        # Check if input is a directory and upload everything
        # including subdirectories.
        if os.path.isdir(path):
            logging.debug("'{}' is a directory.".format(path))
            logging.info("Uploading contents of folder '{}' ...".format(path))
            for root, dirs, files in os.walk(path, topdown=False):
                uploaded_files = []
                for next_path in files + dirs:
                    next_uploaded_files = upload_to_server(
                        workflow, os.path.join(root, next_path), access_token
                    )
                    if next_uploaded_files:
                        uploaded_files.extend(next_uploaded_files)
            return uploaded_files

        # Check if input is an absolute path and upload file.
        else:
            with open(path, "rb") as f:
                fname = os.path.basename(f.name)
                # Calculate the path that will store the file
                # in the workflow controller, by subtracting
                # the workflow root path from the file path
                save_path = path.replace(os.getcwd(), "")
                # Remove prepending dirs named "." or as the upload type
                while len(save_path.split("/")) > 1 and save_path.split("/")[0] == ".":
                    save_path = "/".join(save_path.strip("/").split("/")[1:])
                logging.debug(
                    "'{}' is an absolute filepath.".format(os.path.basename(fname))
                )
                logging.info("Uploading '{}' ...".format(fname))
                try:
                    upload_file(workflow, f, save_path, access_token)
                    logging.info("File '{}' was successfully uploaded.".format(fname))
                    return [save_path]
                except Exception as e:
                    logging.debug(traceback.format_exc())
                    logging.debug(str(e))
                    logging.info(
                        "Something went wrong while uploading {}".format(fname)
                    )
                    raise e


def get_workflow_parameters(workflow, access_token):
    """Get parameters of previously created workflow.

    :param workflow: name or id of the workflow.
    :param access_token: access token of the current user.

    :returns: a dictionary that cointains info about the workflow (``name``, ``type``), and
              a dictionary of workflow parameters under the ``parameters`` key.
    """
    try:
        response, http_response = current_rs_api_client.api.get_workflow_parameters(
            workflow_id_or_name=workflow, access_token=access_token
        ).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Workflow parameters could not be retrieved: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def get_workflow_specification(workflow, access_token):
    """Get specification of previously created workflow.

    :param workflow: name or id of the workflow.
    :param access_token: access token of the current user.

    :returns: a dictionary that cointains two top-level keys: ``parameters``, and
              ``specification`` (which contains a dictionary created from the workflow specification).
    """
    try:
        response, http_response = current_rs_api_client.api.get_workflow_specification(
            workflow_id_or_name=workflow, access_token=access_token
        ).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Workflow specification could not be retrieved: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def delete_workflow(workflow, all_runs: bool, workspace: bool, access_token: str):
    """Delete a workflow.

    Please note that the workspace will always be deleted, even if ``workspace`` is set to ``False``.

    :param workflow: name or id of the workflow.
    :param all_runs: whether to delete all the runs of the workflow.
    :param workspace: whether to delete the workspace of the workflow.
    :param access_token: access token of the current user.

    :return: a dictionary that cointains info about the deleted workflow (``workflow_id``, ``workflow_name``,
             ``status``, ``user``), and a ``message`` key.
    """
    if not workspace:
        logging.warning(
            "Parameter workspace=False was specified in delete_workflow() but workspace will still be deleted."
        )
    workspace = True

    try:
        parameters = {
            "all_runs": all_runs,
            "workspace": workspace,
        }
        response, http_response = current_rs_api_client.api.set_workflow_status(
            workflow_id_or_name=workflow,
            status="deleted",
            access_token=access_token,
            parameters=parameters,
        ).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Workflow run could not be deleted: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def stop_workflow(workflow, force_stop, access_token):
    """Stop a workflow.

    :param workflow: name or id of the workflow.
    :param force_stop: whether to stop the workflow immediately, without
        waiting for the jobs to finish.
    :param access_token: access token of the current user.

    :return: a dictionary that cointains info about the stopped workflow (``workflow_id``, ``workflow_name``,
             ``status``, ``user``), and a ``message`` key.
    """
    try:
        parameters = {"force_stop": force_stop}
        response, http_response = current_rs_api_client.api.set_workflow_status(
            workflow_id_or_name=workflow,
            status="stop",
            access_token=access_token,
            parameters=parameters,
        ).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )
    except HTTPError as e:
        logging.debug(
            "Workflow run could not be stopped: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def diff_workflows(workflow_id_a, workflow_id_b, brief, access_token, context_lines):
    """Return the list of differences between two workflows.

    :param workflow_id_a: UUID which identifies the first workflow.
    :param workflow_id_b: UUID which identifies the second workflow.
    :param brief: Flag specifying desired detail in diff.
    :param context_lines: Optional parameter to set the number of
                          context lines shown in the diff output.
    :param access_token: API token of user requesting diff.

    :return: a list of dictionaries composed by ``asset``, ``type``, ``lines``,
        ``a`` and ``b``. Asset refers to the workflow asset where a
        difference was found, type refers to the asset type, lines refer
        to the lines of the file where the differences are and a, b fields
        are the actual lines that differ.
    """
    try:
        response, http_response = current_rs_api_client.api.get_workflow_diff(
            workflow_id_or_name_a=workflow_id_a,
            workflow_id_or_name_b=workflow_id_b,
            brief=brief,
            context_lines=context_lines,
            access_token=access_token,
        ).result()

        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "File list could not be retrieved: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def open_interactive_session(
    workflow, access_token, interactive_session_type, interactive_session_configuration
):
    """Open an interactive session inside the workflow workspace.

    :param workflow: name or id of the workflow whose workspace will be available inside the
        interactive session.
    :param access_token: Workflow owner REANA access token.
    :param interactive_session_type: Type of interactive session to spawn.
    :param interactive_session_configuration: Specific configuration for
        the interactive session.

    :return: the relative path to the interactive session.
    """
    try:
        response, http_response = current_rs_api_client.api.open_interactive_session(
            workflow_id_or_name=workflow,
            access_token=access_token,
            interactive_session_type=interactive_session_type,
            interactive_session_configuration=interactive_session_configuration,
        ).result()
        if http_response.status_code == 200:
            return response["path"]
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )
    except HTTPError as e:
        logging.debug(
            "Interactive session could not be opened: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def close_interactive_session(workflow, access_token):
    """Close an interactive workflow session.

    :param workflow: name or id of the workflow to close.
    :param access_token: workflow owner REANA access token.

    :return: the relative path to the interactive session.
    """
    try:
        response, http_response = current_rs_api_client.api.close_interactive_session(
            workflow_id_or_name=workflow,
            access_token=access_token,
        ).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )
    except HTTPError as e:
        logging.debug(
            "Interactive session could not be closed: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def mv_files(source, target, workflow, access_token):
    """Move target file(s) within workspace.

    :param source: source filename or path.
    :param target: target filename or path.
    :param workflow: name or id of the workflow.
    :param access_token: token of user.

    :return: a dictionary containing the ``workflow_id``, ``workflow_name``,
             and a ``message`` about the success of the operation.
    """
    try:
        response, http_response = current_rs_api_client.api.move_files(
            source=source,
            target=target,
            workflow_id_or_name=workflow,
            access_token=access_token,
        ).result()

        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Files move command failed: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def get_workflow_disk_usage(workflow, parameters, access_token):
    """Display disk usage workflow.

    :param workflow: name or id of the workflow.
    :param parameters: a dictionary to customize the response. It has the following
        (optional) keys:

        - ``summarize``: a boolean value to indicate whether to summarize the response
          to include only the total workspace disk usage
        - ``search``: a string to filter the response by file name

    :param access_token: access token of the current user.

    :return: a dictionary containing the ``workflow_id``, ``workflow_name``, and the ``user`` ID, with
             a ``disk_usage_info`` keys that contains a list of dictionaries, each of one corresponding
             to a file, with the ``name`` and ``size`` keys.
    """
    try:
        response, http_response = current_rs_api_client.api.get_workflow_disk_usage(
            workflow_id_or_name=workflow,
            parameters=parameters,
            access_token=access_token,
        ).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Workflow disk usage could not be retrieved: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def add_secrets(secrets, overwrite, access_token):
    """Add new secrets.

    :param secrets: dictionary containing all the secrets to be sent.
        The dictionary has the secret names for keys and for each key there is
        a dictionary with two fields:

        - ``value``:  a base64 encoded file or literal string
        - ``type``: ``"file"`` or ``"env"``

    :param overwrite: whether secrets should be overwritten when they
     already exist.
    :param access_token: access token of the current user.

    :return: a dictionary containing the ``message`` key with a success message.
    """
    try:
        response, http_response = current_rs_api_client.api.add_secrets(
            secrets=secrets, access_token=access_token, overwrite=overwrite
        ).result()
        if http_response.status_code == 201:
            return response
        else:
            raise Exception(
                "Expected status code 201 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Secrets could not be added: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        if e.status_code == 409:
            raise REANASecretAlreadyExists()
        else:
            raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def delete_secrets(secrets, access_token):
    """Delete a list of secrets.

    :param secrets: list of secret names to be deleted.
    :param access_token: access token of the current user.

    :return: a list with the names of the deleted secrets.
    """
    try:
        response, http_response = current_rs_api_client.api.delete_secrets(
            secrets=secrets, access_token=access_token
        ).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        if e.response.status_code == 404:
            raise REANASecretDoesNotExist(e.response.json())
        else:
            logging.debug(
                "Secrets could not be deleted: "
                "\nStatus: {}\nReason: {}\n"
                "Message: {}".format(
                    e.response.status_code,
                    e.response.reason,
                    e.response.json()["message"],
                )
            )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def list_secrets(access_token):
    """List user secrets.

    :param access_token: access token of the current user.

    :return: a list of dictionaries, each of one corresponding to a secret, with the
             ``name`` and ``type`` keys.
    """
    try:
        response, http_response = current_rs_api_client.api.get_secrets(
            access_token=access_token
        ).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Secrets could not be listed: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def info(access_token):
    """List general information about the cluster.

    :param access_token: access token of the current user.

    :return: a dictionary containing relevant values and configuration options about the cluster.
             Each key contains a dictionary with the ``title`` key, explaining the meaning of the
             value, and the ``value`` key, containing the value itself.
             Example of the returned keys include ``compute_backends``, ``default_kubernetes_memory_limit``,
             and ``maximum_interactive_session_inactivity_period``.
    """
    try:
        response, http_response = current_rs_api_client.api.info(
            access_token=access_token
        ).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Workspaces could not be listed: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])
    except Exception as e:
        raise e


def get_workflow_retention_rules(workflow, access_token):
    """Get the retention rules of a workflow.

    :param workflow: name or id of the workflow.
    :param access_token: access token of the current user.

    :return: a dictionary containing the ``workflow_id``, ``workflow_name``, and
             the ``retention_rules`` key with a list of dictionaries representing
             the retention rules of the workflow. Each dictionary contains info
             about the affected workspace files, and the schedule of the retention
             rule.
    """
    try:
        (
            response,
            http_response,
        ) = current_rs_api_client.api.get_workflow_retention_rules(
            workflow_id_or_name=workflow,
            access_token=access_token,
        ).result()
        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Workflow retention rules could not be retrieved: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])


def prune_workspace(workflow, include_inputs, include_outputs, access_token):
    """Prune workspace files.

    :param workflow: name or id of the workflow.
    :param include_inputs: whether to also delete inputs.
    :param include_outputs: whether to also delete outputs.
    :param access_token: access token of the current user.

    :return: a dictionary containing the ``workflow_id``, ``workflow_name``, and
             a ``message`` key with the result of the operation.
    """
    try:
        response, http_response = current_rs_api_client.api.prune_workspace(
            workflow_id_or_name=workflow,
            include_inputs=include_inputs,
            include_outputs=include_outputs,
            access_token=access_token,
        ).result()

        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                "{status_code}".format(status_code=http_response.status_code)
            )

    except HTTPError as e:
        logging.debug(
            "Workspace could not be pruned: "
            "\nStatus: {}\nReason: {}\n"
            "Message: {}".format(
                e.response.status_code, e.response.reason, e.response.json()["message"]
            )
        )
        raise Exception(e.response.json()["message"])


def share_workflow(
    workflow, user_email_to_share_with, access_token, message=None, valid_until=None
):
    """Share a workflow with a user.

    :param workflow: name or id of the workflow.
    :param user_email_to_share_with: user to share the workflow with.
    :param access_token: access token of the current user.
    :param message: Optional message to include when sharing the workflow.
    :param valid_until: Specify the date when access to the workflow will expire (format: YYYY-MM-DD).

    :return: a dictionary containing the ``workflow_id``, ``workflow_name``, and
             a ``message`` key with the result of the operation.
    """
    try:
        share_params = {
            "user_email_to_share_with": user_email_to_share_with,
        }

        if message:
            share_params["message"] = message

        if valid_until:
            share_params["valid_until"] = valid_until

        response, http_response = current_rs_api_client.api.share_workflow(
            workflow_id_or_name=workflow,
            share_details=share_params,
            access_token=access_token,
        ).result()

        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                f"{http_response.status_code}"
            )

    except HTTPError as e:
        logging.debug(
            "Workflow could not be shared: "
            f"\nStatus: {e.response.status_code}\nReason: {e.response.reason}\n"
            f"Message: {e.response.json()['message']}"
        )
        raise Exception(e.response.json()["message"])


def unshare_workflow(workflow, user_email_to_unshare_with, access_token):
    """Unshare a workflow with a user.

    :param workflow: name or id of the workflow.
    :param user_email_to_unshare_with: user to unshare the workflow with.
    :param access_token: access token of the current user.

    :return: a dictionary containing the ``workflow_id``, ``workflow_name``, and
             a ``message`` key with the result of the operation.
    """
    try:
        unshare_params = {
            "workflow_id_or_name": workflow,
            "user_email_to_unshare_with": user_email_to_unshare_with,
            "access_token": access_token,
        }

        response, http_response = current_rs_api_client.api.unshare_workflow(
            **unshare_params
        ).result()

        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                f"{http_response.status_code}"
            )

    except HTTPError as e:
        logging.debug(
            "Workflow could not be unshared: "
            f"\nStatus: {e.response.status_code}\nReason: {e.response.reason}\n"
            f"Message: {e.response.json()['message']}"
        )
        raise Exception(e.response.json()["message"])


def get_workflow_sharing_status(workflow, access_token):
    """Get the share status of a workflow.

    :param workflow: name or id of the workflow.
    :param access_token: access token of the current user.

    :return: a dictionary containing the ``workflow_id``, ``workflow_name``, and
             a ``sharing_status`` key with the result of the operation.
    """
    try:
        (
            response,
            http_response,
        ) = current_rs_api_client.api.get_workflow_share_status(
            workflow_id_or_name=workflow,
            access_token=access_token,
        ).result()

        if http_response.status_code == 200:
            return response
        else:
            raise Exception(
                "Expected status code 200 but replied with "
                f"{http_response.status_code}"
            )

    except HTTPError as e:
        logging.debug(
            "Workflow sharing status could not be retrieved: "
            f"\nStatus: {e.response.status_code}\nReason: {e.response.reason}\n"
            f"Message: {e.response.json()['message']}"
        )
        raise Exception(e.response.json()["message"])



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client workflow related commands."""

import json
import logging
import os
import sys
import time
import traceback

import click
import yaml
from jsonschema.exceptions import ValidationError
from reana_client.cli.files import get_files, upload_files
from reana_client.cli.utils import (
    add_access_token_options,
    add_access_token_options_not_required,
    add_pagination_options,
    add_workflow_option,
    check_connection,
    display_formatted_output,
    format_session_uri,
    get_formatted_progress,
    human_readable_or_raw_option,
    key_value_to_dict,
    parse_filter_parameters,
    requires_environments,
    retrieve_workflow_logs,
    follow_workflow_logs,
)
from reana_client.config import (
    ERROR_MESSAGES,
    RUN_STATUSES,
    TIMECHECK,
    CLI_LOGS_FOLLOW_DEFAULT_INTERVAL,
)
from reana_client.printer import display_message
from reana_client.utils import (
    get_reana_yaml_file_path,
    get_workflow_duration,
    get_workflow_name_and_run_number,
    get_workflow_status_change_msg,
    is_uuid_v4,
    load_validate_reana_spec,
    workflow_uuid_or_name,
)
from reana_client.validation.utils import (
    validate_input_parameters,
    validate_workflow_name_parameter,
)
from reana_commons.config import INTERACTIVE_SESSION_TYPES, REANA_COMPUTE_BACKENDS
from reana_commons.errors import REANAValidationError
from reana_commons.validation.operational_options import validate_operational_options


@click.group(help="Workflow management commands")
@click.pass_context
def workflow_management_group(ctx):
    """Top level wrapper for workflow management."""
    logging.debug(ctx.info_name)


@click.group(help="Workflow execution commands")
@click.pass_context
def workflow_execution_group(ctx):
    """Top level wrapper for execution related interaction."""
    logging.debug(ctx.info_name)


@click.group(help="Workflow sharing commands")
@click.pass_context
def workflow_sharing_group(ctx):
    """Top level wrapper for workflow sharing."""
    logging.debug(ctx.info_name)


@workflow_management_group.command("list")
@click.option(
    "-w",
    "--workflow",
    default=None,
    help="List all runs of the given workflow.",
)
@click.option(
    "-s", "--sessions", is_flag=True, help="List all open interactive sessions."
)
@click.option(
    "--format",
    "_format",
    multiple=True,
    help="Format output according to column titles or column values. "
    "Use `<columm_name>=<column_value>` format. "
    "E.g. display workflow with failed status and named test_workflow "
    "`--format status=failed,name=test_workflow`.",
)
@click.option(
    "--json",
    "output_format",
    flag_value="json",
    default=None,
    help="Get output in JSON format.",
)
@click.option(
    "--all",
    "show_all",
    is_flag=True,
    default=False,
    help="Show all workflows including deleted ones.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Print out extra information: workflow id, user id, disk usage, "
    "progress, duration.",
)
@human_readable_or_raw_option
@click.option(
    "--sort",
    "sort_column_name",
    default="CREATED",
    help="Sort the output by specified column",
)
@click.option(
    "--filter",
    "filters",
    multiple=True,
    help="Filter workflow that contains certain filtering criteria. "
    "Use `--filter <columm_name>=<column_value>` pairs. "
    "Available filters are ``name`` and ``status``.",
)
@click.option(
    "--include-duration",
    "include_duration",
    is_flag=True,
    default=False,
    help="Include the duration of the workflows in seconds. In case a workflow is in "
    "progress, its duration as of now will be shown.",
)
@click.option(
    "--include-progress",
    "include_progress",
    is_flag=True,
    default=None,
    help="Include progress information of the workflows.",
)
@click.option(
    "--include-workspace-size",
    "include_workspace_size",
    is_flag=True,
    default=None,
    help="Include size information of the workspace.",
)
@click.option(
    "--show-deleted-runs",
    "show_deleted_runs",
    is_flag=True,
    default=False,
    help="Include deleted workflows in the output.",
)
@click.option(
    "--shared",
    "shared",
    is_flag=True,
    default=False,
    help="List all shared (owned and unowned) workflows.",
)
@click.option(
    "--shared-by",
    "shared_by",
    default=None,
    help="List workflows shared by the specified user.",
)
@click.option(
    "--shared-with",
    "shared_with",
    default=None,
    help="List workflows shared with the specified user.",
)
@add_access_token_options
@add_pagination_options
@check_connection
@click.pass_context
def workflows_list(  # noqa: C901
    ctx,
    workflow,
    sessions,
    _format,
    output_format,
    access_token,
    show_all,
    verbose,
    human_readable_or_raw,
    sort_column_name,
    page,
    size,
    filters,
    include_duration: bool,
    include_progress,
    include_workspace_size,
    show_deleted_runs: bool,
    shared,
    shared_by,
    shared_with,
):  # noqa: D301
    """List all workflows and sessions.

    The ``list`` command lists workflows and sessions. By default, the list of
    workflows is returned. If you would like to see the list of your open
    interactive sessions, you need to pass the ``--sessions`` command-line
    option. If you would like to see the list of all workflows, including those
    shared with you, you need to pass the ``--shared`` command-line option.

    Along with specific user emails, you can pass the following special values
    to the ``--shared-by`` and ``--shared-with`` command-line options:\n
    \t - ``--shared-by anybody``: list workflows shared with you by anybody.\n
    \t - ``--shared-with anybody``: list your shared workflows exclusively.\n
    \t - ``--shared-with nobody``: list your unshared workflows exclusively.\n
    \t - ``--shared-with bob@cern.ch``: list workflows shared with bob@cern.ch

    Examples:\n
    \t $ reana-client list --all\n
    \t $ reana-client list --sessions\n
    \t $ reana-client list --verbose --bytes\n
    \t $ reana-client list --shared\n
    \t $ reana-client list --shared-by bob@cern.ch\n
    \t $ reana-client list --shared-with anybody
    """
    from reana_client.api.client import get_workflows

    if shared_by and shared_with:
        display_message(
            "Please provide either --shared-by or --shared-with, not both.",
            msg_type="error",
        )
        sys.exit(1)

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))
    type = "interactive" if sessions else "batch"

    status_filter = RUN_STATUSES.copy()
    if not show_deleted_runs and not show_all:
        status_filter.remove("deleted")

    search_filter = None
    if filters:
        filter_names = ["name", "status"]
        provided_status_filter, search_filter = parse_filter_parameters(
            filters, filter_names
        )
        if provided_status_filter:
            status_filter = provided_status_filter

    try:
        response = get_workflows(
            access_token,
            type,
            verbose=bool(verbose),
            page=page,
            size=size,
            status=status_filter,
            search=search_filter,
            include_progress=include_progress,
            include_workspace_size=include_workspace_size,
            workflow=workflow,
            shared=shared,
            shared_by=shared_by,
            shared_with=shared_with,
        )
        verbose_headers = ["id", "user"]
        workspace_size_header = ["size"]
        progress_header = ["progress"]
        duration_header = ["duration"]
        headers = {
            "batch": [
                "name",
                "run_number",
                "created",
                "started",
                "ended",
                "status",
            ],
            "interactive": [
                "name",
                "run_number",
                "created",
                "session_type",
                "session_uri",
                "session_status",
            ],
        }
        if verbose:
            headers[type] += verbose_headers
        if verbose or include_workspace_size:
            headers[type] += workspace_size_header
        if verbose or include_progress:
            headers[type] += progress_header
        if verbose or include_duration:
            headers[type] += duration_header

        if shared:
            headers[type] += ["shared_with", "shared_by"]
        else:
            if shared_with:
                headers[type] += ["shared_with"]
            if shared_by:
                headers[type] += ["shared_by"]

        data = []
        for workflow in response:
            name, run_number = get_workflow_name_and_run_number(workflow["name"])
            workflow["name"] = name
            workflow["run_number"] = run_number
            workflow["duration"] = get_workflow_duration(workflow)
            if type == "interactive":
                workflow["session_uri"] = format_session_uri(
                    reana_server_url=ctx.obj.reana_server_url,
                    path=workflow["session_uri"],
                    access_token=access_token,
                )
            row = []
            for header in headers[type]:
                value = None
                if header in progress_header:
                    value = get_formatted_progress(workflow.get("progress"))
                elif header in ["started", "ended"]:
                    _key = (
                        "run_started_at" if header == "started" else "run_finished_at"
                    )
                    value = workflow.get("progress", {}).get(_key) or "-"
                if header == "shared_by":
                    value = workflow.get("owner_email")
                if header == "shared_with":
                    value = ", ".join(workflow.get("shared_with", []))
                if not value:
                    value = workflow.get(header)
                row.append(value)
            data.append(row)

        sort_column_id = 2
        sort_column_name = sort_column_name.lower()
        if sort_column_name in headers[type]:
            sort_column_id = headers[type].index(sort_column_name)

        # Sort by given column, making sure that `None` is at the bottom of the list.
        def get_sort_key(x, column_id):
            if headers[type][column_id] == "run_number":
                return list(map(int, x[column_id].split(".")))
            elif headers[type][column_id] in workspace_size_header:
                return x[column_id]["raw"]
            return x[column_id] is not None, x[column_id]

        data = sorted(
            data,
            key=lambda x: get_sort_key(x, sort_column_id),
            reverse=True,
        )

        for row in data:
            for i, value in enumerate(row):
                # Substitute `None` with "-"
                content = value or "-"
                # Replace raw size with human-readable size if requested
                if headers[type][i] in workspace_size_header:
                    content = value.get(human_readable_or_raw)
                row[i] = content

        workflow_ids = ["{0}.{1}".format(w[0], w[1]) for w in data]
        if os.getenv("REANA_WORKON", "") in workflow_ids:
            active_workflow_idx = workflow_ids.index(os.getenv("REANA_WORKON", ""))
            data[active_workflow_idx][headers[type].index("run_number")] += " *"

        display_formatted_output(data, headers[type], _format, output_format)

    except Exception as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message(
            "Workflow list could not be retrieved: \n{}".format(str(e)),
            msg_type="error",
        )
        sys.exit(1)


@workflow_management_group.command("create")
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True, resolve_path=True),
    default=get_reana_yaml_file_path,
    help="REANA specification file describing the workflow to "
    "execute. [default=reana.yaml]",
)
@click.option(
    "-n",
    "--name",
    "-w",
    "--workflow",
    default="",
    callback=validate_workflow_name_parameter,
    help='Optional name of the workflow. [default is "workflow"]',
)
@click.option(
    "--skip-validation",
    is_flag=True,
    help="If set, specifications file is not validated before "
    "submitting it's contents to REANA server.",
)
@add_access_token_options
@check_connection
@click.pass_context
def workflow_create(ctx, file, name, skip_validation, access_token):  # noqa: D301
    """Create a new workflow.

    The ``create`` command allows to create a new workflow from reana.yaml
    specifications file. The file is expected to be located in the current
    working directory, or supplied via command-line -f option, see examples
    below.

    Examples:\n
    \t $ reana-client create\n
    \t $ reana-client create -w myanalysis\n
    \t $ reana-client create -w myanalysis -f myreana.yaml\n
    """
    from reana_client.api.client import create_workflow
    from reana_client.utils import get_api_url

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))

    # Check that name is not an UUIDv4.
    # Otherwise it would mess up `--workflow` flag usage because no distinction
    # could be made between the name and actual UUID of workflow.
    if is_uuid_v4(name):
        display_message("Workflow name cannot be a valid UUIDv4", msg_type="error")
        sys.exit(1)

    specification_filename = click.format_filename(file)

    try:
        reana_specification = load_validate_reana_spec(
            specification_filename,
            access_token=access_token,
            skip_validation=skip_validation,
            server_capabilities=True,
        )
        logging.info("Connecting to {0}".format(get_api_url()))
        response = create_workflow(reana_specification, name, access_token)
        workflow_name = response["workflow_name"]

        click.echo(click.style(workflow_name, fg="green"))
        # check if command is called from wrapper command
        if "invoked_by_subcommand" in ctx.parent.__dict__:
            ctx.parent.workflow_name = workflow_name
    except Exception as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message(
            "Cannot create workflow {}: \n{}".format(name, str(e)), msg_type="error"
        )
        sys.exit(1)

    # upload specification file by default
    ctx.invoke(
        upload_files,
        workflow=workflow_name,
        filenames=(specification_filename,),
        access_token=access_token,
    )


@workflow_execution_group.command("start")
@add_workflow_option
@add_access_token_options
@check_connection
@click.option(
    "-p",
    "--parameter",
    "parameters",
    multiple=True,
    callback=key_value_to_dict,
    help="Additional input parameters to override "
    "original ones from reana.yaml. "
    "E.g. -p myparam1=myval1 -p myparam2=myval2.",
)
@click.option(
    "-o",
    "--option",
    "options",
    multiple=True,
    callback=key_value_to_dict,
    help="Additional operational options for the workflow execution. "
    "E.g. CACHE=off. (workflow engine - serial) "
    "E.g. --debug (workflow engine - cwl)",
)
@click.option(
    "--follow",
    "follow",
    is_flag=True,
    default=False,
    help="If set, follows the execution of the workflow until termination.",
)
@click.pass_context
def workflow_start(
    ctx, workflow, access_token, parameters, options, follow
):  # noqa: D301
    """Start previously created workflow.

    The ``start`` command allows to start previously created workflow. The
    workflow execution can be further influenced by passing input prameters
    using ``-p`` or ``--parameters`` flag and by setting additional operational
    options using ``-o`` or ``--options``.  The input parameters and operational
    options can be repetitive. For example, to disable caching for the Serial
    workflow engine, you can set ``-o CACHE=off``.

    Examples:\n
    \t $ reana-client start -w myanalysis.42 -p sleeptime=10 -p myparam=4\n
    \t $ reana-client start -w myanalysis.42 -p myparam1=myvalue1 -o CACHE=off
    """
    from reana_client.api.client import (
        get_workflow_parameters,
        get_workflow_status,
        start_workflow,
    )
    from reana_client.utils import get_api_url

    def display_status(workflow: str, current_status: str):
        """Display the current status of the workflow."""
        status_msg = get_workflow_status_change_msg(workflow, current_status)
        if current_status in ["deleted", "failed", "stopped"]:
            display_message(status_msg, msg_type="error")
            sys.exit(1)
        else:
            display_message(status_msg, msg_type="success")

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))

    parsed_parameters = {"input_parameters": parameters, "operational_options": options}
    if workflow:
        if parameters or options:
            try:
                response = get_workflow_parameters(workflow, access_token)
                workflow_type = response["type"]
                original_parameters = response["parameters"]
                validate_operational_options(
                    workflow_type, parsed_parameters["operational_options"]
                )

                parsed_parameters["input_parameters"] = validate_input_parameters(
                    parsed_parameters["input_parameters"], original_parameters
                )
            except REANAValidationError as e:
                display_message(e.message, msg_type="error")
                sys.exit(1)
            except Exception as e:
                display_message(
                    "Could not apply given input parameters: "
                    "{0} \n{1}".format(parameters, str(e)),
                    msg_type="error",
                )
                sys.exit(1)
        try:
            logging.info("Connecting to {0}".format(get_api_url()))
            response = start_workflow(workflow, access_token, parsed_parameters)
            current_status = get_workflow_status(workflow, access_token).get("status")
            display_status(workflow, current_status)

            if follow:
                # keep printing the current status of the workflow
                while current_status in ["pending", "queued", "running"]:
                    time.sleep(TIMECHECK)
                    current_status = get_workflow_status(workflow, access_token).get(
                        "status"
                    )
                    display_status(workflow, current_status)

                    if current_status == "finished":
                        display_message(
                            "Listing workflow output files...", msg_type="info"
                        )
                        ctx.invoke(
                            get_files,
                            workflow=workflow,
                            access_token=access_token,
                            output_format="url",
                            human_readable_or_raw="raw",
                        )
        except Exception as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            display_message(
                "Cannot start workflow {}: \n{}".format(workflow, str(e)),
                msg_type="error",
            )
            sys.exit(1)


@workflow_execution_group.command("restart")
@add_workflow_option
@add_access_token_options
@check_connection
@click.option(
    "-p",
    "--parameter",
    "parameters",
    multiple=True,
    callback=key_value_to_dict,
    help="Additional input parameters to override "
    "original ones from reana.yaml. "
    "E.g. -p myparam1=myval1 -p myparam2=myval2.",
)
@click.option(
    "-o",
    "--option",
    "options",
    multiple=True,
    callback=key_value_to_dict,
    help="Additional operational options for the workflow execution. "
    "E.g. CACHE=off. (workflow engine - serial) "
    "E.g. --debug (workflow engine - cwl)",
)
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True, resolve_path=True),
    help="REANA specification file describing the workflow to "
    "execute. [default=reana.yaml]",
)
@click.pass_context
def workflow_restart(
    ctx, workflow, access_token, parameters, options, file
):  # noqa: D301
    """Restart previously run workflow.

    The ``restart`` command allows to restart a previous workflow on the same
    workspace.

    Note that workflow restarting can be used in a combination with operational
    options ``FROM`` and ``TARGET``. You can also pass a modified workflow
    specification with ``-f`` or ``--file`` flag.

    You can furthermore use modified input prameters using ``-p`` or
    ``--parameters`` flag and by setting additional operational options using
    ``-o`` or ``--options``.  The input parameters and operational options can be
    repetitive.

    Examples:\n
    \t $ reana-client restart -w myanalysis.42 -p sleeptime=10 -p myparam=4\n
    \t $ reana-client restart -w myanalysis.42 -p myparam=myvalue\n
    \t $ reana-client restart -w myanalysis.42 -o TARGET=gendata\n
    \t $ reana-client restart -w myanalysis.42 -o FROM=fitdata
    """
    from reana_client.api.client import (
        get_workflow_parameters,
        get_workflow_status,
        start_workflow,
    )
    from reana_client.utils import get_api_url

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))

    parsed_parameters = {
        "input_parameters": parameters,
        "operational_options": options,
        "restart": True,
    }
    if file:
        specification_filename = click.format_filename(file)
        parsed_parameters["reana_specification"] = load_validate_reana_spec(
            click.format_filename(file)
        )
        # upload new specification
        ctx.invoke(
            upload_files,
            workflow=workflow,
            filenames=(specification_filename,),
            access_token=access_token,
        )

    if parameters or options:
        try:
            if "reana_specification" in parsed_parameters:
                workflow_type = parsed_parameters["reana_specification"]["workflow"][
                    "type"
                ]
                original_parameters = (
                    parsed_parameters["reana_specification"]
                    .get("inputs", {})
                    .get("parameters", {})
                )
            else:
                response = get_workflow_parameters(workflow, access_token)
                workflow_type = response["type"]
                original_parameters = response["parameters"]

            parsed_parameters["operational_options"] = validate_operational_options(
                workflow_type, parsed_parameters["operational_options"]
            )
            parsed_parameters["input_parameters"] = validate_input_parameters(
                parsed_parameters["input_parameters"], original_parameters
            )

        except REANAValidationError as e:
            display_message(e.message, msg_type="error")
            sys.exit(1)
        except Exception as e:
            display_message(
                "Could not apply given input parameters: "
                "{0} \n{1}".format(parameters, str(e)),
                msg_type="error",
            )
            sys.exit(1)

    try:
        logging.info("Connecting to {0}".format(get_api_url()))
        response = start_workflow(workflow, access_token, parsed_parameters)
        workflow = response["workflow_name"] + "." + str(response["run_number"])
        current_status = get_workflow_status(workflow, access_token).get("status")
        display_message(
            get_workflow_status_change_msg(workflow, current_status),
            msg_type="success",
        )
    except Exception as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message(
            "Cannot start workflow {}: \n{}".format(workflow, str(e)),
            msg_type="error",
        )
        sys.exit(1)


@workflow_execution_group.command("status")
@add_workflow_option
@click.option(
    "--format",
    "_format",
    multiple=True,
    help="Format output by displaying only certain columns. "
    "E.g. --format name,status.",
)
@click.option(
    "--json",
    "output_format",
    flag_value="json",
    default=None,
    help="Get output in JSON format.",
)
@click.option(
    "--include-duration",
    "include_duration",
    is_flag=True,
    default=False,
    help="Include the duration of the workflow in seconds. In case the workflow is in "
    "progress, its duration as of now will be shown.",
)
@add_access_token_options
@check_connection
@click.option("-v", "--verbose", count=True, help="Set status information verbosity.")
@click.pass_context
def workflow_status(  # noqa: C901
    ctx, workflow, _format, output_format, include_duration, access_token, verbose
):  # noqa: D301
    """Get status of a workflow.

    The ``status`` command allow to retrieve status of a workflow. The status can
    be created, queued, running, failed, etc. You can increase verbosity or
    filter retrieved information by passing appropriate command-line options.

    Examples:\n
    \t $ reana-client status -w myanalysis.42\n
    \t $ reana-client status -w myanalysis.42 -v --json
    """
    from reana_client.api.client import get_workflow_status

    def render_progress(finished_jobs, total_jobs):
        if total_jobs:
            return "{0}/{1}".format(finished_jobs, total_jobs)
        else:
            return "-/-"

    def add_data_from_response(row, data, headers):
        name, run_number = get_workflow_name_and_run_number(row["name"])
        total_jobs = row["progress"].get("total")
        if total_jobs:
            total_jobs = total_jobs.get("total")
        else:
            total_jobs = 0
        finished_jobs = row["progress"].get("finished")
        if finished_jobs:
            finished_jobs = finished_jobs.get("total")
        else:
            finished_jobs = 0

        parsed_response = list(
            map(str, [name, run_number, row["created"], row["status"]])
        )
        if row["progress"]["total"].get("total") or 0 > 0:
            if "progress" not in headers:
                headers += ["progress"]
                parsed_response.append(render_progress(finished_jobs, total_jobs))

        if row["status"] in ["running", "finished", "failed", "stopped"]:
            started_at = row["progress"].get("run_started_at")
            finished_at = row["progress"].get("run_finished_at")
            if started_at:
                after_created_pos = headers.index("created") + 1
                headers.insert(after_created_pos, "started")
                parsed_response.insert(after_created_pos, started_at)
                if finished_at:
                    after_started_pos = headers.index("started") + 1
                    headers.insert(after_started_pos, "ended")
                    parsed_response.insert(after_started_pos, finished_at)

        data.append(parsed_response)
        return data

    def add_verbose_data_from_response(response, verbose_headers, headers, data):
        for k in verbose_headers:
            if k == "command":
                current_command = response["progress"]["current_command"]
                if current_command:
                    if current_command.startswith('bash -c "cd '):
                        current_command = current_command[
                            current_command.index(";") + 2 : -2
                        ]
                    data[-1] += [current_command]
                else:
                    if "current_step_name" in response["progress"] and response[
                        "progress"
                    ].get("current_step_name"):
                        current_step_name = response["progress"].get(
                            "current_step_name"
                        )
                        data[-1] += [current_step_name]
                    else:
                        headers.remove("command")
            else:
                data[-1] += [response.get(k)]
        return data

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))
    try:
        workflow_response = get_workflow_status(workflow, access_token)
        headers = ["name", "run_number", "created", "status"]
        verbose_headers = ["id", "user", "command"]
        data = []
        add_data_from_response(workflow_response, data, headers)
        if verbose:
            headers += verbose_headers
            add_verbose_data_from_response(
                workflow_response, verbose_headers, headers, data
            )
        if verbose or include_duration:
            headers += ["duration"]
            data[-1] += [get_workflow_duration(workflow_response) or "-"]

        display_formatted_output(data, headers, _format, output_format)

    except Exception as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message(
            "Cannot retrieve the status of a workflow {}: \n"
            "{}".format(workflow, str(e)),
            msg_type="error",
        )
        sys.exit(1)


@workflow_execution_group.command("logs")
@add_workflow_option
@click.option("--json", "json_format", count=True, help="Get output in JSON format.")
@add_access_token_options
@click.option(
    "--filter",
    "filters",
    multiple=True,
    help="Filter job logs to include only those steps that match certain filtering criteria. Use --filter name=value pairs. Available filters are compute_backend, docker_img, status and step.",
)
@click.option(
    "--follow",
    "follow",
    is_flag=True,
    default=False,
    help="Follow the logs of a running workflow or job (similar to tail -f).",
)
@click.option(
    "-i",
    "--interval",
    "interval",
    default=CLI_LOGS_FOLLOW_DEFAULT_INTERVAL,
    help=f"Sleep time in seconds between log polling if log following is enabled. [default={CLI_LOGS_FOLLOW_DEFAULT_INTERVAL}]",
)
@add_pagination_options
@check_connection
@click.pass_context
def workflow_logs(
    ctx,
    workflow,
    access_token,
    json_format,
    follow,
    interval,
    filters=None,
    page=None,
    size=None,
):  # noqa: D301
    """Get workflow logs.

    The ``logs`` command allows to retrieve logs of a running workflow.

    Examples:\n
    \t $ reana-client logs -w myanalysis.42\n
    \t $ reana-client logs -w myanalysis.42 --json\n
    \t $ reana-client logs -w myanalysis.42 --filter status=running\n
    \t $ reana-client logs -w myanalysis.42 --filter step=myfit --follow\n
    """
    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))

    if json_format and follow:
        display_message(
            "Ignoring --json as it cannot be used together with --follow.",
            msg_type="warning",
        )

    available_filters = {
        "step": "job_name",
        "compute_backend": "compute_backend",
        "docker_img": "docker_img",
        "status": "status",
    }
    steps = []
    chosen_filters = dict()

    if filters:
        try:
            for f in filters:
                key, value = f.split("=")
                if key not in available_filters:
                    display_message(
                        "Filter '{}' is not valid.\n"
                        "Available filters are '{}'.".format(
                            key,
                            "' '".join(sorted(available_filters.keys())),
                        ),
                        msg_type="error",
                    )
                    sys.exit(1)
                elif key == "step":
                    steps.append(value)
                else:
                    # Case insensitive for compute backends
                    if (
                        key == "compute_backend"
                        and value.lower() in REANA_COMPUTE_BACKENDS
                    ):
                        value = REANA_COMPUTE_BACKENDS[value.lower()]
                    elif key == "status" and value not in RUN_STATUSES:
                        display_message(
                            "Input status value {} is not valid. ".format(value),
                            msg_type="error",
                        ),
                        sys.exit(1)
                    chosen_filters[key] = value
        except Exception as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            display_message(
                "Please provide complete --filter name=value pairs, "
                "for example --filter status=running.\n"
                "Available filters are '{}'.".format(
                    "' '".join(sorted(available_filters.keys()))
                ),
                msg_type="error",
            )
            sys.exit(1)

    try:
        if follow:
            follow_workflow_logs(workflow, access_token, interval, steps)
        else:
            retrieve_workflow_logs(
                workflow,
                access_token,
                json_format,
                filters,
                steps,
                chosen_filters,
                available_filters,
                page,
                size,
            )
    except Exception as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message(
            "Cannot retrieve logs for workflow {}: \n{}".format(workflow, str(e)),
            msg_type="error",
        )
        sys.exit(1)


@workflow_execution_group.command("validate")
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True, resolve_path=True),
    default=get_reana_yaml_file_path,
    help="REANA specification file describing the workflow to "
    "execute. [default=reana.yaml]",
)
@click.option(
    "--environments",
    is_flag=True,
    default=False,
    help="If set, check all runtime environments specified in REANA "
    "specification file. [default=False]",
)
@click.option(
    "--pull",
    is_flag=True,
    default=False,
    callback=requires_environments,
    help="If set, try to pull remote environment image from registry to perform "
    "validation locally. Requires ``--environments`` flag. [default=False]",
)
@click.option(
    "--server-capabilities",
    is_flag=True,
    default=False,
    help="If set, check the server capabilities such as workspace validation. "
    "[default=False]",
)
@add_access_token_options_not_required
@click.pass_context
def workflow_validate(
    ctx, file, environments, pull, server_capabilities, access_token
):  # noqa: D301
    """Validate workflow specification file.

    The ``validate`` command allows to check syntax and validate the reana.yaml
    workflow specification file.

    Examples:\n
    \t $ reana-client validate -f reana.yaml
    """
    if server_capabilities:
        if access_token:
            check_connection(lambda: None)()
        else:
            display_message(ERROR_MESSAGES["missing_access_token"], msg_type="error")
            ctx.exit(1)
    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))
    try:
        load_validate_reana_spec(
            click.format_filename(file),
            access_token=access_token,
            skip_validate_environments=not environments,
            pull_environment_image=pull,
            server_capabilities=server_capabilities,
        )

    except (ValidationError, REANAValidationError) as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message(
            "{0} is not a valid REANA specification:\n{1}".format(
                click.format_filename(file), e.message
            ),
            msg_type="error",
        )
        sys.exit(1)
    except yaml.parser.ParserError as e:
        logging.debug(traceback.format_exc())
        display_message(
            "{0} is not a valid YAML file:\n{1}".format(
                click.format_filename(file, shorten=True), e
            ),
            msg_type="error",
        )
        sys.exit(1)
    except Exception as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message(
            "Something went wrong when trying to validate {}".format(file),
            msg_type="error",
        )
        sys.exit(1)


@workflow_execution_group.command("stop")
@click.option(
    "--force",
    "force_stop",
    is_flag=True,
    default=False,
    help="Stop a workflow without waiting for jobs to finish.",
)
@add_workflow_option
@add_access_token_options
@check_connection
@click.pass_context
def workflow_stop(ctx, workflow, force_stop, access_token):  # noqa: D301
    """Stop a running workflow.

    The ``stop`` command allows to hard-stop the running workflow process. Note
    that soft-stopping of the workflow is currently not supported. This command
    should be therefore used with care, only if you are absolutely sure that
    there is no point in continuing the running the workflow.

    Example:\n
    \t $ reana-client stop -w myanalysis.42 --force
    """
    from reana_client.api.client import stop_workflow

    if not force_stop:
        display_message(
            "Graceful stop not implement yet. If you really want to "
            "stop your workflow without waiting for jobs to finish"
            " use: --force option",
            msg_type="error",
        )
        raise click.Abort()

    if workflow:
        try:
            logging.info("Sending a request to stop workflow {}".format(workflow))
            stop_workflow(workflow, force_stop, access_token)
            display_message(
                get_workflow_status_change_msg(workflow, "stopped"),
                msg_type="success",
            )
        except Exception as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            display_message(
                "Cannot stop workflow {}: \n{}".format(workflow, str(e)),
                msg_type="error",
            )
            sys.exit(1)


@workflow_execution_group.command("run")
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True, resolve_path=True),
    default=get_reana_yaml_file_path,
    help="REANA specification file describing the workflow to "
    "execute. [default=reana.yaml]",
)
@click.option(
    "-n",
    "--name",
    "-w",
    "--workflow",
    default="",
    callback=validate_workflow_name_parameter,
    help='Optional name of the workflow. [default is "workflow"]',
)
@click.option(
    "--skip-validation",
    is_flag=True,
    help="If set, specifications file is not validated before "
    "submitting it's contents to REANA server.",
)
@click.option(
    "-p",
    "--parameter",
    "parameters",
    multiple=True,
    callback=key_value_to_dict,
    help="Additional input parameters to override "
    "original ones from reana.yaml. "
    "E.g. -p myparam1=myval1 -p myparam2=myval2.",
)
@click.option(
    "-o",
    "--option",
    "options",
    multiple=True,
    callback=key_value_to_dict,
    help="Additional operational options for the workflow execution. "
    "E.g. CACHE=off.",
)
@click.option(
    "--follow",
    "follow",
    is_flag=True,
    default=False,
    help="If set, follows the execution of the workflow until termination.",
)
@add_access_token_options
@check_connection
@click.pass_context
def workflow_run(
    ctx, file, name, skip_validation, access_token, parameters, options, follow
):  # noqa: D301
    """Shortcut to create, upload, start a new workflow.

    The ``run`` command allows to create a new workflow, upload its input files
    and start it in one command.

    Examples:\n
    \t $ reana-client run -w myanalysis-test-small -p myparam=mysmallvalue\n
    \t $ reana-client run -w myanalysis-test-big -p myparam=mybigvalue
    """
    # set context parameters for subcommand
    ctx.invoked_by_subcommand = True
    ctx.workflow_name = ""
    display_message("Creating a workflow...", msg_type="info")
    ctx.invoke(
        workflow_create,
        file=file,
        name=name,
        skip_validation=skip_validation,
        access_token=access_token,
    )
    display_message("Uploading files...", msg_type="info")
    ctx.invoke(
        upload_files,
        workflow=ctx.workflow_name,
        access_token=access_token,
    )
    display_message("Starting workflow...", msg_type="info")
    ctx.invoke(
        workflow_start,
        workflow=ctx.workflow_name,
        access_token=access_token,
        parameters=parameters,
        options=options,
        follow=follow,
    )


@workflow_management_group.command("delete")
@click.option(
    "--include-all-runs",
    "all_runs",
    is_flag=True,
    help="Delete all runs of a given workflow.",
)
@click.option(
    "--include-workspace",
    "should_delete_workspace",
    is_flag=True,
    help="Delete workspace from REANA.",
)
@add_workflow_option
@add_access_token_options
@check_connection
@click.pass_context
def workflow_delete(
    ctx, workflow: str, all_runs: bool, should_delete_workspace: bool, access_token: str
):  # noqa: D301
    """Delete a workflow.

    The ``delete`` command removes workflow run(s) from the database.
    Note that the workspace and any open session attached to it will always be
    deleted, even when ``--include-workspace`` is not specified.
    Note also that you can remove all past runs of a workflow by specifying ``--include-all-runs`` flag.

    Example:\n
    \t $ reana-client delete -w myanalysis.42\n
    \t $ reana-client delete -w myanalysis.42 --include-all-runs
    """
    from reana_client.api.client import delete_workflow
    from reana_client.utils import get_api_url

    should_delete_workspace = True

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))

    if workflow:
        try:
            logging.info("Connecting to {0}".format(get_api_url()))
            delete_workflow(workflow, all_runs, should_delete_workspace, access_token)
            if all_runs:
                message = "All workflows named '{}' have been deleted.".format(
                    workflow.split(".")[0]
                )
            else:
                message = get_workflow_status_change_msg(workflow, "deleted")
            display_message(message, msg_type="success")

        except Exception as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            display_message(
                "Cannot delete workflow {} \n{}".format(workflow, str(e)),
                msg_type="error",
            )
            sys.exit(1)


@workflow_management_group.command("diff")
@click.argument(
    "workflow_a",
    default=os.environ.get("REANA_WORKON", None),
    callback=workflow_uuid_or_name,
)
@click.argument("workflow_b", callback=workflow_uuid_or_name)
@click.option(
    "-q",
    "--brief",
    is_flag=True,
    help="If not set, differences in the contents of the files in the two "
    "workspaces are shown.",
)
@click.option(
    "-u",
    "-U",
    "--unified",
    "context_lines",
    type=int,
    default=5,
    help="Sets number of context lines for workspace diff output.",
)
@add_access_token_options
@check_connection
@click.pass_context
def workflow_diff(
    ctx, workflow_a, workflow_b, brief, access_token, context_lines
):  # noqa: D301
    """Show diff between two workflows.

    The ``diff`` command allows to compare two workflows, the workflow_a and
    workflow_b, which must be provided as arguments. The output will show the
    difference in workflow run parameters, the generated files, the logs, etc.

    Examples:\n
    \t $ reana-client diff myanalysis.42 myotheranalysis.43\n
    \t $ reana-client diff myanalysis.42 myotheranalysis.43 --brief
    """
    from reana_client.api.client import diff_workflows

    logging.debug("command: {}".format(ctx.command_path.replace(" ", ".")))
    for p in ctx.params:
        logging.debug("{param}: {value}".format(param=p, value=ctx.params[p]))

    def print_color_diff(lines):
        for line in lines:
            line_color = None
            if line[0] == "@":
                line_color = "cyan"
            elif line[0] == "-":
                line_color = "red"
            elif line[0] == "+":
                line_color = "green"
            click.secho(line, fg=line_color)

    leading_mark = "==>"
    try:
        response = diff_workflows(
            workflow_a, workflow_b, brief, access_token, str(context_lines)
        )
        if response.get("reana_specification"):
            specification_diff = json.loads(response["reana_specification"])
            nonempty_sections = {k: v for k, v in specification_diff.items() if v}
            if not nonempty_sections:
                click.secho(
                    "{} No differences in REANA specifications.".format(leading_mark),
                    bold=True,
                    fg="yellow",
                )
            # Rename section workflow -> specification
            if "workflow" in nonempty_sections:
                nonempty_sections["specification"] = nonempty_sections.pop("workflow")
            for section, content in nonempty_sections.items():
                click.secho(
                    "{} Differences in workflow {}".format(leading_mark, section),
                    bold=True,
                    fg="yellow",
                )
                print_color_diff(content)
        display_message("")  # Leave 1 line for separation
        workspace_diff = json.loads(response.get("workspace_listing"))
        if workspace_diff:
            workspace_diff = workspace_diff.splitlines()
            click.secho(
                "{} Differences in workflow workspace".format(leading_mark),
                bold=True,
                fg="yellow",
            )
            print_color_diff(workspace_diff)

    except Exception as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message(
            "Something went wrong when trying to get diff:\n{}".format(str(e)),
            msg_type="error",
        )
        sys.exit(1)


@click.group(help="Workspace interactive commands")
def interactive_group():
    """Workspace interactive commands."""
    pass


@interactive_group.command("open")
@add_workflow_option
@click.argument(
    "interactive-session-type",
    metavar="interactive-session-type",
    default=INTERACTIVE_SESSION_TYPES[0],
    type=click.Choice(INTERACTIVE_SESSION_TYPES),
)
@click.option(
    "-i",
    "--image",
    help="Docker image which will be used to spawn the interactive session. "
    "Overrides the default image for the selected type.",
)
@add_access_token_options
@check_connection
@click.pass_context
def workflow_open_interactive_session(
    ctx, workflow, interactive_session_type, image, access_token
):  # noqa: D301
    """Open an interactive session inside the workspace.

    The ``open`` command allows to open interactive session processes on top of
    the workflow workspace, such as Jupyter notebooks. This is useful to
    quickly inspect and analyse the produced files while the workflow is still
    running.

    Examples:\n
    \t $ reana-client open -w myanalysis.42 jupyter
    """
    from reana_client.api.client import info, open_interactive_session

    if workflow:
        try:
            logging.info("Opening an interactive session on {}".format(workflow))
            interactive_session_configuration = {
                "image": image or None,
            }
            path = open_interactive_session(
                workflow,
                access_token,
                interactive_session_type,
                interactive_session_configuration,
            )
            display_message(
                "Interactive session opened successfully", msg_type="success"
            )
            click.secho(
                format_session_uri(
                    reana_server_url=ctx.obj.reana_server_url,
                    path=path,
                    access_token=access_token,
                ),
                fg="green",
            )
            display_message(
                "It could take several minutes to start the interactive session."
            )
            reana_info = info(access_token)
            max_inactivity_days_entry = (
                reana_info.get("maximum_interactive_session_inactivity_period") or {}
            )
            max_inactivity_days = max_inactivity_days_entry.get("value")
            if max_inactivity_days:
                display_message(
                    f"Please note that it will be automatically closed after {max_inactivity_days} days of inactivity."
                )
        except Exception as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            display_message(
                "Interactive session could not be opened: \n{}".format(str(e)),
                msg_type="error",
            )
            sys.exit(1)
    else:
        display_message("Cannot find workflow {}".format(workflow), msg_type="error")


@interactive_group.command("close")
@add_workflow_option
@add_access_token_options
@check_connection
def workflow_close_interactive_session(workflow, access_token):  # noqa: D301
    """Close an interactive session.

    The ``close`` command allows to shut down any interactive sessions that you
    may have running. You would typically use this command after you finished
    exploring data in the Jupyter notebook and after you have transferred any
    code created in your interactive session.

    Examples:\n
    \t $ reana-client close -w myanalysis.42
    """
    from reana_client.api.client import close_interactive_session

    if workflow:
        try:
            logging.info("Closing an interactive session on {}".format(workflow))
            close_interactive_session(workflow, access_token)
            display_message(
                "Interactive session for workflow {}"
                " was successfully closed".format(workflow),
                msg_type="success",
            )
        except Exception as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            display_message(
                "Interactive session could not be closed: \n{}".format(str(e)),
                msg_type="error",
            )
            sys.exit(1)
    else:
        display_message("Cannot find workflow {} ".format(workflow), msg_type="error")


@workflow_sharing_group.command("share-add")
@check_connection
@add_workflow_option
@add_access_token_options
@click.option(
    "-u",
    "--user",
    "users",
    multiple=True,
    help="Users to share the workflow with.",
    required=True,
)
@click.option(
    "-m",
    "--message",
    help="Optional message that is sent to the user(s) with the sharing invitation.",
)
@click.option(
    "--valid-until",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    help="Optional date when access to the workflow will expire for the given user(s) (format: YYYY-MM-DD).",
)
@click.pass_context
def workflow_share_add(
    ctx, workflow, access_token, users, message, valid_until
):  # noqa D412
    """Share a workflow with other users (read-only).

    The `share-add` command allows sharing a workflow with other users. The
    users will be able to view the workflow but not modify it.

    Examples:

    \t $ reana-client share-add -w myanalysis.42 --user bob@example.org

    \t $ reana-client share-add -w myanalysis.42 --user bob@example.org --user cecile@example.org --message "Please review my analysis" --valid-until 2025-12-31
    """
    from reana_client.api.client import share_workflow

    share_errors = []
    shared_users = []

    if valid_until:
        valid_until = valid_until.strftime("%Y-%m-%d")

    for user in users:
        try:
            logging.info(f"Sharing workflow {workflow} with user {user}")
            share_workflow(
                workflow,
                user,
                access_token,
                message=message,
                valid_until=valid_until,
            )
            shared_users.append(user)
        except Exception as e:
            share_errors.append(f"Failed to share {workflow} with {user}: {str(e)}")
            logging.debug(traceback.format_exc())

    if shared_users:
        display_message(
            f"{workflow} is now read-only shared with {', '.join(shared_users)}",
            msg_type="success",
        )

    for error in share_errors:
        display_message(error, msg_type="error")

    if share_errors:
        sys.exit(1)


@workflow_sharing_group.command("share-remove")
@check_connection
@add_workflow_option
@add_access_token_options
@click.option(
    "-u",
    "--user",
    "users",
    multiple=True,
    help="Users to unshare the workflow with.",
    required=True,
)
@click.pass_context
def share_workflow_remove(ctx, workflow, access_token, users):  # noqa D412
    """Unshare a workflow.

    The `share-remove` command allows for unsharing a workflow. The workflow
    will no longer be visible to the users with whom it was shared.

    Example:

        $ reana-client share-remove -w myanalysis.42 --user bob@example.org
    """
    from reana_client.api.client import unshare_workflow

    unshare_errors = []
    unshared_users = []

    for user in users:
        try:
            logging.info(f"Unsharing workflow {workflow} with user {user}")
            unshare_workflow(workflow, user, access_token)
            unshared_users.append(user)
        except Exception as e:
            unshare_errors.append(f"Failed to unshare {workflow} with {user}: {str(e)}")
            logging.debug(traceback.format_exc())

    if unshared_users:
        display_message(
            f"{workflow} is no longer shared with {', '.join(unshared_users)}",
            msg_type="success",
        )
    if unshare_errors:
        for error in unshare_errors:
            display_message(error, msg_type="error")
        sys.exit(1)


@workflow_sharing_group.command("share-status")
@check_connection
@add_workflow_option
@add_access_token_options
@click.option(
    "--format",
    "format_",
    multiple=True,
    default=None,
    help="Format output according to column titles or column "
    "values. Use <columm_name>=<column_value> format.",
)
@click.option(
    "--json",
    "output_format",
    flag_value="json",
    default=None,
    help="Get output in JSON format.",
)
@click.pass_context
def share_workflow_status(
    ctx, workflow, format_, output_format, access_token
):  # noqa D412
    """Show with whom a workflow is shared.

    The `share-status` command allows for checking with whom a workflow is
    shared.

    Example:

        $ reana-client share-status -w myanalysis.42
    """
    from reana_client.api.client import get_workflow_sharing_status

    try:
        sharing_status = get_workflow_sharing_status(workflow, access_token)
    except Exception as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message(
            "An error occurred while checking workflow sharing status:\n{}".format(
                str(e)
            ),
            msg_type="error",
        )
        sys.exit(1)

    shared_with = sharing_status.get("shared_with", [])
    if shared_with:
        headers = ["user_email", "valid_until"]
        data = [
            [
                entry["user_email"],
                (entry["valid_until"] if entry["valid_until"] is not None else "-"),
            ]
            for entry in shared_with
        ]

        display_formatted_output(data, headers, format_, output_format)
    else:
        display_message(
            f"Workflow {workflow} is not shared with anyone.", msg_type="info"
        )



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2022 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client retention-rules related commands."""

import logging
import sys
from typing import Optional, Tuple

import click

from reana_client.api.client import get_workflow_retention_rules
from reana_client.cli.utils import (
    add_access_token_options,
    add_workflow_option,
    check_connection,
    display_formatted_output,
)
from reana_client.printer import display_message


@click.group(help="Workspace file retention commands")
def retention_rules_group():
    """Workspace file retention commands."""
    pass


@retention_rules_group.command()
@add_workflow_option
@add_access_token_options
@check_connection
@click.option(
    "--format",
    "_format",
    multiple=True,
    help="Format output according to column titles or column values. "
    "Use `<columm_name>=<column_value>` format. "
    "E.g. display pattern and status of active retention rules "
    "`--format workspace_files,status=active`.",
)
@click.option(
    "--json",
    "output_format",
    flag_value="json",
    default=None,
    help="Get output in JSON format.",
)
def retention_rules_list(
    access_token: str, workflow: str, _format: Tuple[str], output_format: Optional[str]
) -> None:  # noqa: D301
    """List the retention rules for a workflow.

    Example:\n
    \t $ reana-client retention-rules-list -w myanalysis.42
    """
    try:
        rules = get_workflow_retention_rules(workflow, access_token).get(
            "retention_rules", []
        )
    except Exception as e:
        logging.debug(e, exc_info=True)
        display_message(str(e), msg_type="error")
        sys.exit(1)

    sorted_rules = sorted(rules, key=lambda rule: rule["retention_days"])

    headers = ["workspace_files", "retention_days", "apply_on", "status"]
    rows = [[rule[h] or "-" for h in headers] for rule in sorted_rules]
    display_formatted_output(rows, headers, _format, output_format)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2024 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client test commands."""

import sys
import click
import logging
import traceback
import time
from reana_client.cli.utils import add_access_token_options, check_connection
from reana_commons.gherkin_parser.parser import (
    parse_and_run_tests,
    AnalysisTestStatus,
)
from reana_commons.gherkin_parser.data_fetcher import DataFetcherBase
from reana_commons.gherkin_parser.errors import FeatureFileError
from reana_client.printer import display_message
from reana_client.api.client import (
    list_files,
    get_workflow_disk_usage,
    get_workflow_logs,
    get_workflow_status,
    get_workflow_specification,
    download_file,
)
from reana_client.cli.utils import add_workflow_option


class DataFetcherClient(DataFetcherBase):
    """Implementation of the DataFetcherBase using reana_client.api.client methods."""

    def __init__(self, access_token):
        """Initialize DataFetcherClient with access token."""
        self.access_token = access_token

    def list_files(self, workflow, file_name=None, page=None, size=None, search=None):
        """Return the list of files for a given workflow workspace."""
        return list_files(workflow, self.access_token, file_name, page, size, search)

    def get_workflow_disk_usage(self, workflow, parameters):
        """Display disk usage workflow."""
        return get_workflow_disk_usage(workflow, parameters, self.access_token)

    def get_workflow_logs(self, workflow, steps=None, page=None, size=None):
        """Get logs from a workflow engine, use existing API function."""
        return get_workflow_logs(workflow, self.access_token, steps, page, size)

    def get_workflow_status(self, workflow):
        """Get of a previously created workflow."""
        return get_workflow_status(workflow, self.access_token)

    def get_workflow_specification(self, workflow):
        """Get specification of previously created workflow."""
        return get_workflow_specification(workflow, self.access_token)

    def download_file(self, workflow, file_path):
        """Download the requested file if it exists."""
        return download_file(workflow, file_path, self.access_token)


@click.group(help="Workflow run test commands")
def test_group():
    """Workflow run test commands."""


@test_group.command("test")
@click.option(
    "-n",
    "--test-files",
    multiple=True,
    default=None,
    help="Gherkin file for testing properties of a workflow execution. Overrides files in reana.yaml if provided.",
)
@click.pass_context
@add_access_token_options
@check_connection
@add_workflow_option
def test(ctx, workflow, test_files, access_token):
    r"""
    Test workflow execution, based on a given Gherkin file.

    Gherkin files can be specified in the reana specification file (reana.yaml),
    or by using the ``-n`` option.

    The ``test`` command allows for testing of a workflow execution,
    by assessing whether it meets certain properties specified in a
    chosen gherkin file.

    Example:
        $ reana-client test -w myanalysis -n test_analysis.feature
        $ reana-client test -w myanalysis
        $ reana-client test -w myanalysis -n test1.feature -n test2.feature
    """
    start_time = time.time()
    try:
        workflow_status = get_workflow_status(
            workflow=workflow, access_token=access_token
        )
        status = workflow_status["status"]
        workflow_name = workflow_status["name"]
    except Exception as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message(f"Could not find workflow ``{workflow}``.", msg_type="error")
        sys.exit(1)

    if status != "finished":
        display_message(
            f"``{workflow}`` is {status}. It must be finished to run tests.",
            msg_type="error",
        )
        sys.exit(1)

    if not test_files:
        reana_specification = get_workflow_specification(workflow, access_token)
        try:
            test_files = reana_specification["specification"]["tests"]["files"]
        except KeyError as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            display_message(
                "No test files specified in reana.yaml and no -n option provided.",
                msg_type="error",
            )
            sys.exit(1)

    passed = 0
    failed = 0
    data_fetcher = DataFetcherClient(access_token)
    for test_file in test_files:
        click.echo("\n", nl=False)
        display_message(f'Testing file "{test_file}"...', msg_type="info")
        try:
            results = parse_and_run_tests(test_file, workflow_name, data_fetcher)
        except FileNotFoundError as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            display_message(f"Test file {test_file} not found.", msg_type="error")
            sys.exit(1)
        except FeatureFileError as e:
            logging.debug(traceback.format_exc())
            logging.debug(str(e))
            display_message(
                f"Error parsing feature file {test_file}: {e}", msg_type="error"
            )
            sys.exit(1)

        for scenario in results[1]:
            if scenario.result == AnalysisTestStatus.failed:
                display_message(
                    f'Scenario "{scenario.scenario}"', msg_type="error", indented=True
                )
                failed += 1
            else:
                display_message(
                    f'Scenario "{scenario.scenario}"', msg_type="success", indented=True
                )
                passed += 1

    end_time = time.time()
    duration = round(end_time - start_time)
    click.echo(f"\n{passed} passed, {failed} failed in {duration}s")
    if failed > 0:
        sys.exit(1)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024, 2025, 2026 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client"""

from __future__ import absolute_import, print_function

import os
import re

from setuptools import find_packages, setup

readme = open("README.md").read()
history = open("CHANGELOG.md").read()

extras_require = {
    "docs": [
        "myst-parser",
        "Sphinx>=1.5.1",
        "sphinx-rtd-theme>=0.1.9",
        "sphinx-click>=1.0.4",
    ],
    "tests": [
        "pytest-reana>=0.95.0a6,<0.96.0",
        "reana-commons[kubernetes]>=0.95.0a14,<0.96.0",
    ],
}

extras_require["all"] = []
for key, reqs in extras_require.items():
    if ":" == key[0]:
        continue
    extras_require["all"].extend(reqs)

install_requires = [
    "click>=7",
    "pathspec>=0.9.0,<1.0 ; python_version<'3.9'",
    "pathspec>=0.9.0 ; python_version>='3.9'",
    "jsonpointer>=2.0",
    "reana-commons[yadage,snakemake,cwl]>=0.95.0a14,<0.96.0",
    "tablib>=0.12.1,<0.13",
    "werkzeug>=0.14.1 ; python_version<'3.10'",
    "werkzeug>=0.15.0 ; python_version>='3.10'",
]

packages = find_packages()


# Get the version string. Cannot be done with import!
with open(os.path.join("reana_client", "version.py"), "rt") as f:
    version = re.search(r'__version__\s*=\s*"(?P<version>.*)"\n', f.read()).group(
        "version"
    )

setup(
    name="reana-client",
    version=version,
    description=__doc__,
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    author="REANA",
    author_email="info@reana.io",
    url="https://github.com/reanahub/reana-client",
    packages=[
        "reana_client",
    ],
    zip_safe=False,
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "reana-client = reana_client.cli:cli",
            "reana-cwl-runner = reana_client.cli.cwl_runner:cwl_runner",
        ],
    },
    python_requires=">=3.8",
    extras_require=extras_require,
    install_requires=install_requires,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client utils."""

import base64
from datetime import datetime
import logging
import os
import pathlib
import sys
import traceback
from typing import Dict, Optional
from uuid import UUID

from reana_commons.utils import get_workflow_status_change_verb
from reana_commons.specification import load_reana_spec

from reana_client.config import reana_yaml_valid_file_names
from reana_client.printer import display_message
from reana_client.validation.utils import validate_reana_spec


def workflow_uuid_or_name(ctx, param, value):
    """Get UUID of workflow from configuration / cache file based on name."""
    if not value:
        display_message(
            "Workflow name must be provided either with "
            "`--workflow` option or with REANA_WORKON "
            "environment variable",
            msg_type="error",
        )
        sys.exit(1)
    else:
        return value


def load_validate_reana_spec(
    filepath,
    access_token=None,
    skip_validation=False,
    skip_validate_environments=True,
    pull_environment_image=False,
    server_capabilities=False,
):
    """Load and validate reana specification file.

    :raises IOError: Error while reading REANA spec file from given `filepath`.
    :raises ValidationError: Given REANA spec file does not validate against
        REANA specification.
    """

    try:
        reana_yaml = load_reana_spec(filepath)
        validate_reana_spec(
            reana_yaml,
            filepath,
            access_token=access_token,
            skip_validation=skip_validation,
            skip_validate_environments=skip_validate_environments,
            pull_environment_image=pull_environment_image,
            server_capabilities=server_capabilities,
        )

        if reana_yaml["workflow"]["type"] == "yadage":
            # We don't send the loaded Yadage workflow spec to the cluster as
            # it may result in inconsistencies between what's displayed in the
            # UI and the actual spec loaded at the workflow engine level.
            # More info: https://github.com/reanahub/reana-client/pull/462#discussion_r585794297
            reana_yaml["workflow"]["specification"] = None

        return reana_yaml
    except Exception as e:
        raise e


def is_uuid_v4(uuid_or_name):
    """Check if given string is a valid UUIDv4."""
    # Based on https://gist.github.com/ShawnMilo/7777304
    try:
        uuid = UUID(uuid_or_name, version=4)
    except Exception:
        return False

    return uuid.hex == uuid_or_name.replace("-", "")


def is_regular_path(path: str) -> bool:
    """Check if path does not refer to a symbolic link."""
    full_path = pathlib.Path(os.path.abspath(path))
    for parent in full_path.parents:
        if not parent.is_dir() or parent.is_symlink():
            return False
    return not full_path.is_symlink()


def get_workflow_name_and_run_number(workflow_name):
    """Return name and run_number of a workflow.

    :param workflow_name: String representing Workflow name.

        Name might be in format 'reana.workflow.123' with arbitrary
        number of dot-delimited substrings, where last substring specifies
        the run number of the workflow this workflow name refers to.

        If name does not contain a valid run number, name without run number
        is returned.
    """
    # Try to split a dot-separated string.
    try:
        name, run_number = workflow_name.split(".", 1)

        try:
            float(run_number)
        except ValueError:
            # `workflow_name` was split, so it is a dot-separated string
            # but it didn't contain a valid `run_number`.
            # Assume that this dot-separated string is the name of
            # the workflow and return just this without a `run_number`.
            return workflow_name, ""

        return name, run_number

    except ValueError:
        # Couldn't split. Probably not a dot-separated string.
        # Return the name given as parameter without a `run_number`.
        return workflow_name, ""


def get_workflow_duration(workflow: Dict) -> Optional[int]:
    """Calculate the duration of the workflow.

    :param workflow: Workflow details returned by the server.
    :return: The duration of the workflow in seconds or ``None`` if the starting
        time is not present.
    """

    progress = workflow.get("progress", {})
    run_started_at = progress.get("run_started_at")
    run_finished_at = progress.get("run_finished_at")
    run_stopped_at = progress.get("run_stopped_at")

    duration = None
    if run_started_at:
        start_time = datetime.fromisoformat(run_started_at)
        if run_finished_at:
            end_time = datetime.fromisoformat(run_finished_at)
        elif run_stopped_at:
            end_time = datetime.fromisoformat(run_stopped_at)
        else:
            end_time = datetime.utcnow()
        duration = round((end_time - start_time).total_seconds())
    return duration


def get_workflow_root():
    """Return the current workflow root directory."""
    reana_yaml = get_reana_yaml_file_path()
    workflow_root = os.getcwd()
    while True:
        file_list = os.listdir(workflow_root)
        parent_dir = os.path.dirname(workflow_root)
        if reana_yaml in file_list:
            break
        else:
            if workflow_root == parent_dir:
                display_message(
                    "Not a workflow directory (or any of the parent directories).\n"
                    "Please upload from inside the directory containing "
                    "the reana.yaml file of your workflow.",
                    msg_type="error",
                )
                sys.exit(1)
            else:
                workflow_root = parent_dir
    workflow_root += "/"
    return workflow_root


def get_workflow_status_change_msg(workflow, status):
    """Choose verb conjugation depending on status.

    :param workflow: Workflow name whose status changed.
    :param status: String which represents the status the workflow changed to.
    """
    return "{workflow} {verb} {status}".format(
        workflow=workflow, verb=get_workflow_status_change_verb(status), status=status
    )


def parse_secret_from_literal(literal):
    """Parse a literal string, into a secret dict.

    :param literal: String containg a key and a value. (e.g. 'KEY=VALUE')
    :returns secret: Dictionary in the format suitable for sending
    via http request.
    """
    try:
        key, value = literal.split("=", 1)
        secret = {
            key: {
                "value": base64.b64encode(value.encode("utf-8")).decode("utf-8"),
                "type": "env",
            }
        }
        return secret

    except ValueError as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message(
            'Option "{0}" is invalid: \n'
            'For literal strings use "SECRET_NAME=VALUE" format'.format(literal),
            msg_type="error",
        )
        sys.exit(1)


def parse_secret_from_path(path):
    """Parse a file path into a secret dict.

    :param path: Path of the file containing secret
    :returns secret: Dictionary in the format suitable for sending
     via http request.
    """
    try:
        with open(os.path.expanduser(path), "rb") as file:
            file_name = os.path.basename(path)
            secret = {
                file_name: {
                    "value": base64.b64encode(file.read()).decode("utf-8"),
                    "type": "file",
                }
            }
        return secret
    except FileNotFoundError as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        display_message(
            "File {0} could not be uploaded: {0} does not exist.".format(path),
            msg_type="error",
        )
        sys.exit(1)


def get_api_url():
    """Obtain REANA server API URL."""
    server_url = os.getenv("REANA_SERVER_URL")
    return server_url.strip(" \t\n\r/") if server_url else None


def get_reana_yaml_file_path():
    """REANA specification file location."""
    matches = [path for path in reana_yaml_valid_file_names if os.path.exists(path)]
    if len(matches) == 0:
        display_message(
            "No REANA specification file (reana.yaml) found. Exiting.",
            msg_type="error",
        )
        sys.exit(1)
    if len(matches) > 1:
        display_message(
            "Found {0} REANA specification files ({1}). "
            "Please use only one. Exiting.".format(len(matches), ", ".join(matches)),
            msg_type="error",
        )
        sys.exit(1)
    for path in reana_yaml_valid_file_names:
        if os.path.exists(path):
            return path
    # If none of the valid paths exists, fall back to reana.yaml.
    return "reana.yaml"



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024, 2026 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Version information for REANA-Client.

This file is imported by ``reana_client.__init__`` and parsed by
``setup.py``.
"""

from __future__ import absolute_import, print_function

__version__ = "0.95.0a4"



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2022 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client workspace validation."""

import sys
from typing import List, Optional

from reana_commons.errors import REANAValidationError
from reana_commons.validation.utils import validate_workspace

from reana_client.printer import display_message


def _validate_workspace(
    root_path: str, available_workspaces: Optional[List[str]]
) -> None:
    """Validate workspace in REANA specification file.

    :param root_path: workspace root path to be validated.
    :param available_workspaces: a list of the available workspaces.

    :raises ValidationError: Given workspace in REANA spec file does not validate against
        allowed workspaces.
    """
    if root_path:
        display_message(
            "Verifying workspace in REANA specification file...",
            msg_type="info",
        )
        try:
            validate_workspace(root_path, available_workspaces)
            display_message(
                "Workflow workspace appears valid.",
                msg_type="success",
                indented=True,
            )
        except REANAValidationError as e:
            display_message(e.message, msg_type="error")
            sys.exit(1)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2024 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client test tests."""

from click.testing import CliRunner
from reana_client.cli import cli
from unittest.mock import patch

from reana_commons.gherkin_parser.parser import AnalysisTestStatus, TestResult
from reana_commons.gherkin_parser.errors import FeatureFileError
from dataclasses import replace

passed_test = TestResult(
    scenario="scenario1",
    error_log=None,
    result=AnalysisTestStatus.passed,
    failed_testcase=None,
    feature="Run Duration",
    checked_at="2024-01-01T00:00:00.000000",
)

failed_test = TestResult(
    scenario="scenario2",
    error_log="Test designed to fail",
    result=AnalysisTestStatus.failed,
    failed_testcase="Scenario to fail",
    feature="Run Duration",
    checked_at="2024-01-01T00:00:00.000000",
)


def test_test_workflow_not_found():
    """Test test command when workflow is not found."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolated_filesystem():
        result = runner.invoke(
            cli,
            [
                "test",
                "-w",
                "myanalysis",
                "-n",
                "test_analysis.feature",
                "-t",
                "000000",
            ],
        )
    assert "Could not find workflow ``myanalysis``." in result.output
    assert result.exit_code == 1


@patch(
    "reana_client.cli.test.get_workflow_status",
    return_value={"status": "running", "name": "myanalysis"},
)
def test_test_workflow_not_finished(mock_get_workflow_status):
    """Test test command when workflow is not finished."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        result = runner.invoke(
            cli,
            [
                "test",
                "-w",
                "myanalysis",
                "-n",
                "test_analysis.feature",
                "-t",
                "000000",
            ],
        )
        assert (
            "``myanalysis`` is running. It must be finished to run tests."
            in result.output
        )
        assert result.exit_code == 1


@patch(
    "reana_client.cli.test.get_workflow_status",
    return_value={"status": "deleted", "name": "myanalysis"},
)
def test_test_workflow_deleted(mock_get_workflow_status):
    """Test test command when workflow is deleted."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        result = runner.invoke(
            cli,
            ["test", "-w", "myanalysis", "-n", "test_analysis.feature", "-t", "000000"],
        )
        assert (
            "``myanalysis`` is deleted. It must be finished to run tests."
            in result.output
        )
        assert result.exit_code == 1


@patch(
    "reana_client.cli.test.get_workflow_status",
    return_value={"status": "finished", "name": "myanalysis"},
)
@patch(
    "reana_client.cli.test.get_workflow_specification",
    return_value={"specification": {"inputs": {"directories": ["data"]}}},
)
def test_test_no_test_files(mock_get_workflow_status, mock_get_workflow_specification):
    """Test test command when no test files are specified."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        result = runner.invoke(cli, ["test", "-w", "myanalysis", "-t", "000000"])
        assert (
            "No test files specified in reana.yaml and no -n option provided."
            in result.output
        )
        assert result.exit_code == 1


@patch(
    "reana_client.cli.test.get_workflow_status",
    return_value={"status": "finished", "name": "myanalysis"},
)
@patch(
    "reana_client.cli.test.get_workflow_specification",
    return_value={"specification": {"inputs": {"directories": ["data"]}}},
)
def test_test_no_test_files_with_test_file_option(
    mock_get_workflow_status, mock_get_workflow_specification
):
    """Test test command when no test files are specified in reana.yml and when the test file option is provided."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        result = runner.invoke(
            cli,
            [
                "test",
                "-w",
                "myanalysis",
                "-t",
                "000000",
                "-n",
                "test_analysis.feature",
            ],
        )
        assert 'Testing file "test_analysis.feature"' in result.output


@patch(
    "reana_client.cli.test.get_workflow_status",
    return_value={"status": "finished", "name": "myanalysis"},
)
@patch(
    "reana_client.cli.test.get_workflow_specification",
    return_value={
        "specification": {
            "tests": {"files": ["test_analysis.feature", "test_analysis2.feature"]}
        }
    },
)
def test_test_multiple_test_files_with_test_file_option(
    mock_get_workflow_status, mock_get_workflow_specification
):
    """Test test command when multiple test files are specified in reana.yml and test file option is provided.
    In this case, the test-file option should be used instead of the test files specified in reana.yml.
    """
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        result = runner.invoke(
            cli,
            [
                "test",
                "-w",
                "myanalysis",
                "-t",
                "000000",
                "-n",
                "use_this.feature",
            ],
        )
        assert 'Testing file "use_this.feature"' in result.output
        assert 'Testing file "test_analysis.feature"' not in result.output
        assert 'Testing file "test_analysis2.feature"' not in result.output


@patch(
    "reana_client.cli.test.get_workflow_status",
    return_value={"status": "finished", "name": "myanalysis"},
)
@patch(
    "reana_client.cli.test.get_workflow_specification",
    return_value={
        "specification": {"tests": {"files": ["use-me.feature", "me-too.feature"]}}
    },
)
@patch(
    "reana_client.cli.test.parse_and_run_tests",
    return_value=(
        "myanalysis",
        [passed_test],
    ),
)
def test_test_files_from_spec(
    mock_get_workflow_status, get_workflow_specification, mock_parse_and_run_tests
):
    """Test test command when test files are specified in reana.yml."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        result = runner.invoke(
            cli,
            ["test", "-w", "myanalysis", "-t", "000000"],
        )
        assert 'Testing file "use-me.feature"' in result.output
        assert 'Testing file "me-too.feature"' in result.output


@patch(
    "reana_client.cli.test.get_workflow_status",
    return_value={"status": "finished", "name": "myanalysis"},
)
@patch(
    "reana_client.cli.test.parse_and_run_tests",
    side_effect=FeatureFileError,
)
def test_test_parser_error(mock_get_workflow_status, mock_parse_and_run_tests):
    """Test test command when parser error occurs."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        result = runner.invoke(
            cli,
            [
                "test",
                "-w",
                "myanalysis",
                "-t",
                "000000",
                "-n",
                "test_analysis.feature",
            ],
        )
        assert "Error parsing feature file test_analysis.feature" in result.output
        assert result.exit_code == 1


@patch(
    "reana_client.cli.test.get_workflow_status",
    return_value={"status": "finished", "name": "myanalysis"},
)
@patch(
    "reana_client.cli.test.parse_and_run_tests",
    side_effect=FileNotFoundError,
)
def test_test_feature_file_not_found(
    mock_get_workflow_status, mock_parse_and_run_tests
):
    """Test test command when a feature file is not found."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        result = runner.invoke(
            cli,
            [
                "test",
                "-w",
                "myanalysis",
                "-t",
                "000000",
                "-n",
                "test_analysis.feature",
            ],
        )
        assert "Test file test_analysis.feature not found." in result.output
        assert result.exit_code == 1


@patch(
    "reana_client.cli.test.get_workflow_status",
    return_value={"status": "finished", "name": "myanalysis"},
)
@patch(
    "reana_client.cli.test.parse_and_run_tests",
    return_value=(
        "myanalysis",
        [passed_test],
    ),
)
def test_test_multiple_test_files(mock_workflow_status, mock_parse_and_run_tests):
    """Test test command when multiple test files are specified."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        result = runner.invoke(
            cli,
            [
                "test",
                "-w",
                "myanalysis",
                "-t",
                "000000",
                "-n",
                "test_analysis.feature",
                "-n",
                "test_analysis2.feature",
            ],
        )
        assert 'Testing file "test_analysis.feature"' in result.output
        assert 'Testing file "test_analysis2.feature"' in result.output


@patch(
    "reana_client.cli.test.get_workflow_status",
    return_value={"status": "finished", "name": "myanalysis"},
)
@patch(
    "reana_client.cli.test.parse_and_run_tests",
    return_value=(
        "myanalysis",
        [passed_test, replace(passed_test, scenario="scenario2")],
    ),
)
def test_test_all_scenarios_pass(mock_workflow_status, mock_parse_and_run_tests):
    """Test test command when tests pass."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolated_filesystem():
        result = runner.invoke(
            cli,
            ["test", "-w", "myanalysis", "-n", "test_analysis.feature", "-t", "000000"],
        )
        assert "SUCCESS" in result.output and "ERROR" not in result.output
        assert result.exit_code == 0


@patch(
    "reana_client.cli.test.get_workflow_status",
    return_value={"status": "finished", "name": "myanalysis"},
)
@patch(
    "reana_client.cli.test.parse_and_run_tests",
    return_value=(
        "myanalysis",
        [replace(failed_test, scenario="scenario1"), failed_test],
    ),
)
def test_test_all_scenarios_fail(mock_workflow_status, mock_parse_and_run_tests):
    """Test test command when tests fail."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        result = runner.invoke(
            cli,
            ["test", "-w", "myanalysis", "-n", "test_analysis.feature", "-t", "000000"],
        )
        assert "ERROR" in result.output and "SUCCESS" not in result.output
        assert result.exit_code == 1


@patch(
    "reana_client.cli.test.get_workflow_status",
    return_value={"status": "finished", "name": "myanalysis"},
)
@patch(
    "reana_client.cli.test.parse_and_run_tests",
    return_value=(
        "myanalysis",
        [passed_test, failed_test],
    ),
)
def test_test_some_scenarios_pass(mock_workflow_status, mock_parse_and_run_tests):
    """Test test command when some tests pass and some fail."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolated_filesystem():
        result = runner.invoke(
            cli,
            ["test", "-w", "myanalysis", "-n", "test_analysis.feature", "-t", "000000"],
        )
        assert "SUCCESS" in result.output and "ERROR" in result.output
        assert result.exit_code == 1



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024, 2026 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA command line interface client."""

import warnings

warnings.filterwarnings(
    "ignore", message="pkg_resources is deprecated", module="yadageschemas"
)

import logging  # noqa: E402
import os  # noqa: E402
import sys  # noqa: E402

import click  # noqa: E402
from reana_client.cli import (  # noqa: E402
    files,
    ping,
    quotas,
    retention_rules,
    secrets,
    test,
    workflow,
)
from reana_client.utils import get_api_url  # noqa: E402
from urllib3 import disable_warnings  # noqa: E402

DEBUG_LOG_FORMAT = (
    "[%(asctime)s] p%(process)s "
    "{%(pathname)s:%(lineno)d} "
    "%(levelname)s - %(message)s"
)

LOG_FORMAT = "[%(levelname)s] %(message)s"


class Config(object):
    """Configuration object to share across commands."""

    def __init__(self):
        """Initialize config variables."""
        self.reana_server_url = get_api_url()


class ReanaCLI(click.Group):
    """REANA command line interface."""

    cmd_groups = [
        quotas.quota_group,
        ping.configuration_group,
        workflow.workflow_management_group,
        workflow.workflow_execution_group,
        workflow.workflow_sharing_group,
        workflow.interactive_group,
        files.files_group,
        retention_rules.retention_rules_group,
        secrets.secrets_group,
        test.test_group,
    ]

    def __init__(self, name=None, commands=None, **attrs):
        """Initialize REANA client commands."""
        disable_warnings()
        click.Group.__init__(self, name, **attrs)
        for group in ReanaCLI.cmd_groups:
            for cmd in group.commands.items():
                self.add_command(cmd=cmd[1], name=cmd[0])

    def format_commands(self, ctx, formatter):
        """Overides default click cmd display."""
        if ReanaCLI.cmd_groups:
            max_cmd_length = len(max([max(name) for name in self.list_commands(ctx)]))
            limit = formatter.width - 6 - max_cmd_length
            rows = []
            for group in ReanaCLI.cmd_groups:
                item = {"rows": []}
                item["group_help"] = group.get_short_help_str(limit)
                for command in sorted(group.commands.items()):
                    if command[1] is None:
                        continue
                    if command[1].hidden:
                        continue
                    command_help = command[1].get_short_help_str(limit)
                    item["rows"].append((command[0], command_help))
                rows.append(item)
            for item in rows:
                with formatter.section(item["group_help"]):
                    formatter.write_dl(item["rows"])


@click.command(cls=ReanaCLI)
@click.option(
    "--loglevel",
    "-l",
    help="Sets log level",
    type=click.Choice(["DEBUG", "INFO", "WARNING"]),
    default="WARNING",
)
@click.pass_context
@click.pass_obj
def cli(obj, ctx, loglevel):
    """REANA client for interacting with REANA server."""
    logging.basicConfig(
        format=DEBUG_LOG_FORMAT if loglevel == "DEBUG" else LOG_FORMAT,
        stream=sys.stderr,
        level=loglevel,
    )
    ctx.obj = obj or Config()
